import unittest
import platform

_marker = object()

class TestCatalog(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.catalog import Catalog
        return Catalog

    def _makeOne(self, family=_marker):
        klass = self._getTargetClass()
        if family is _marker:
            return klass()
        return klass(family)

    def test_klass_provides_ICatalog(self):
        klass = self._getTargetClass()
        from zope.interface.verify import verifyClass
        from zerodbext.catalog.catalog import ICatalog
        verifyClass(ICatalog, klass)

    def test_inst_provides_ICatalog(self):
        klass = self._getTargetClass()
        from zope.interface.verify import verifyObject
        from zerodbext.catalog.catalog import ICatalog
        inst = self._makeOne()
        verifyObject(ICatalog, inst)

    def test_ctor_defaults(self):
        from BTrees import family32
        catalog = self._makeOne()
        self.assertTrue(catalog.family is family32)

    def test_ctor_explicit_family(self):
        from BTrees import family64
        catalog = self._makeOne(family64)
        self.assertTrue(catalog.family is family64)

    def test_clear(self):
        catalog = self._makeOne()
        idx = DummyIndex()
        catalog['name'] = idx
        catalog.clear()
        self.assertEqual(idx.cleared, True)

    def test_index_doc(self):
        catalog = self._makeOne()
        idx = DummyIndex()
        catalog['name'] = idx
        catalog.index_doc(1, 'value')
        self.assertEqual(idx.docid, 1)
        self.assertEqual(idx.value, 'value')

    def test_index_doc_nonint_docid(self):
        catalog = self._makeOne()
        idx = DummyIndex()
        catalog['name'] = idx
        self.assertRaises(ValueError, catalog.index_doc, 'abc', 'value')

    def test_reindex_doc(self):
        catalog = self._makeOne()
        idx = DummyIndex()
        catalog['name'] = idx
        catalog.reindex_doc(1, 'value')
        self.assertEqual(idx.reindexed_docid, 1)
        self.assertEqual(idx.reindexed_ob, 'value')

    def test_unindex_doc(self):
        catalog = self._makeOne()
        idx = DummyIndex()
        catalog['name'] = idx
        catalog.unindex_doc(1)
        self.assertEqual(idx.unindexed, 1)

    def test_setitem_guard(self):
        catalog = self._makeOne()
        self.assertRaises(ValueError, catalog.__setitem__, 'a', None)

    def test_search(self):
        from BTrees.IFBTree import IFSet
        catalog = self._makeOne()
        c1 = IFSet([1, 2, 3])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        c2 = IFSet([3, 4, 5])
        idx2 = DummyIndex(c2)
        catalog['name2'] = idx2
        numdocs, result = catalog.search(name1={}, name2={})
        self.assertEqual(numdocs, 1)
        self.assertEqual(numdocs.total, 1)
        self.assertEqual(list(result), [3])

    def test_search_index_returns_empty(self):
        from BTrees.IFBTree import IFSet
        catalog = self._makeOne()
        c1 = IFSet([])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        c2 = IFSet([3, 4, 5])
        idx2 = DummyIndex(c2)
        catalog['name2'] = idx2
        numdocs, result = catalog.search(name1={}, name2={})
        self.assertEqual(numdocs, 0)
        self.assertEqual(numdocs.total, 0)
        self.assertEqual(list(result), [])

    def test_search_no_intersection(self):
        from BTrees.IFBTree import IFSet
        catalog = self._makeOne()
        c1 = IFSet([1, 2])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        c2 = IFSet([3, 4, 5])
        idx2 = DummyIndex(c2)
        catalog['name2'] = idx2
        numdocs, result = catalog.search(name1={}, name2={})
        self.assertEqual(numdocs, 0)
        self.assertEqual(numdocs.total, 0)
        self.assertEqual(list(result), [])

    def test_search_index_query_order_returns_empty(self):
        from BTrees.IFBTree import IFSet
        catalog = self._makeOne()
        c1 = IFSet([1, 2])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        c2 = IFSet([])
        idx2 = DummyIndex(c2)
        catalog['name2'] = idx2
        numdocs, result = catalog.search(name1={}, name2={},
                                         index_query_order=['name2', 'name1'])
        self.assertEqual(numdocs, 0)
        self.assertEqual(numdocs.total, 0)
        self.assertEqual(list(result), [])

    def test_search_no_indexes_in_search(self):
        catalog = self._makeOne()
        numdocs, result = catalog.search()
        self.assertEqual(numdocs, 0)
        self.assertEqual(numdocs.total, 0)
        self.assertEqual(list(result), [])

    def test_search_noindex(self):
        catalog = self._makeOne()
        self.assertRaises(ValueError, catalog.search, name1={})

    def test_search_noindex_ordered(self):
        catalog = self._makeOne()
        self.assertRaises(ValueError, catalog.search, name1={},
                          index_query_order=['name1'])

    def test_apply(self):
        catalog = self._makeOne()
        from BTrees.IFBTree import IFSet
        c1 = IFSet([1, 2, 3])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        c2 = IFSet([3, 4, 5])
        idx2 = DummyIndex(c2)
        catalog['name2'] = idx2
        numdocs, result = catalog.apply({'name1':{}, 'name2':{}})
        self.assertEqual(numdocs, 1)
        self.assertEqual(numdocs.total, 1)
        self.assertEqual(list(result), [3])

    def test_search_with_sortindex_ascending(self):
        catalog = self._makeOne()
        from BTrees.IFBTree import IFSet
        c1 = IFSet([1, 2, 3, 4, 5])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        c2 = IFSet([3, 4, 5])
        idx2 = DummyIndex(c2)
        catalog['name2'] = idx2
        numdocs, result = catalog.search(
            name1={}, name2={}, sort_index='name1')
        self.assertEqual(numdocs, 3)
        self.assertEqual(numdocs.total, 3)
        self.assertEqual(list(result), ['sorted1', 'sorted2', 'sorted3'])

    def test_search_with_sortindex_reverse(self):
        catalog = self._makeOne()
        from BTrees.IFBTree import IFSet
        c1 = IFSet([1, 2, 3, 4, 5])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        c2 = IFSet([3, 4, 5])
        idx2 = DummyIndex(c2)
        catalog['name2'] = idx2
        numdocs, result = catalog.search(
            name1={}, name2={}, sort_index='name1',
            reverse=True)
        self.assertEqual(numdocs, 3)
        self.assertEqual(numdocs.total, 3)
        self.assertEqual(list(result), ['sorted3', 'sorted2', 'sorted1'])

    def test_search_with_sort_type(self):
        catalog = self._makeOne()
        from BTrees.IFBTree import IFSet
        c1 = IFSet([1, 2, 3, 4, 5])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        from zerodbext.catalog.indexes.field import FWSCAN
        numdocs, result = catalog.search(name1={}, sort_index='name1',
                                         limit=1, sort_type=FWSCAN)
        self.assertEqual(idx1.sort_type, FWSCAN)

    def test_limited(self):
        catalog = self._makeOne()
        from BTrees.IFBTree import IFSet
        c1 = IFSet([1, 2, 3, 4, 5])
        idx1 = DummyIndex(c1)
        catalog['name1'] = idx1
        numdocs, result = catalog.search(name1={}, sort_index='name1', limit=1)
        self.assertEqual(numdocs, 1)
        self.assertEqual(numdocs.total, 5)
        self.assertEqual(idx1.limit, 1)

    def _test_functional_merge(self, **extra):
        catalog = self._makeOne()
        from zerodbext.catalog.indexes.field import CatalogFieldIndex
        from zerodbext.catalog.indexes.keyword import CatalogKeywordIndex
        from zerodbext.catalog.indexes.text import CatalogTextIndex
        from zerodbext.catalog.indexes.path2 import CatalogPathIndex2
        class Content(object):
            def __init__(self, field, keyword, text, path):
                self.field = field
                self.keyword = keyword
                self.text = text
                self.path = path
        field = CatalogFieldIndex('field')
        keyword = CatalogKeywordIndex('keyword')
        text = CatalogTextIndex('text')
        path = CatalogPathIndex2('path')
        catalog['field'] = field
        catalog['keyword'] = keyword
        catalog['text'] = text
        catalog['path'] = path
        map = {
            1:Content('field1', ['keyword1', 'same'], 'text one', '/path1'),
            2:Content('field2', ['keyword2', 'same'], 'text two',
                      '/path1/path2'),
            3:Content('field3', ['keyword3', 'same'], 'text three',
                      '/path1/path2/path3'),
            }
        for num, doc in map.items():
            catalog.index_doc(num, doc)
        num, result = catalog.search(field=('field1', 'field1'), **extra)
        self.assertEqual(num, 1)
        self.assertEqual(list(result), [1])
        num, result = catalog.search(field=('field2', 'field2'), **extra)
        self.assertEqual(num, 1)
        self.assertEqual(list(result), [2])
        num, result = catalog.search(field=('field2', 'field2'),
                                     keyword='keyword2', **extra)
        self.assertEqual(num, 1)
        self.assertEqual(list(result), [2])
        num, result = catalog.search(field=('field2', 'field2'), text='two',
                                     **extra)
        self.assertEqual(num, 1)
        self.assertEqual(list(result), [2])
        num, result = catalog.search(text='text', keyword='same', **extra)
        self.assertEqual(num, 3)
        self.assertEqual(list(result), [1,2,3])

        num, result = catalog.search(text='text', path='/path1', **extra)
        self.assertEqual(num, 2)
        self.assertEqual(list(result), [2,3])

    def test_functional_index_merge_unordered(self):
        return self._test_functional_merge()

    def test_functional_index_merge_ordered(self):
        return self._test_functional_merge(
            index_query_order=['field', 'keyword', 'text', 'path'])

class TestFileStorageCatalogFactory(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.catalog import FileStorageCatalogFactory
        return FileStorageCatalogFactory

    def _makeOne(self, filename, appname, **kw):
        klass = self._getTargetClass()
        return klass(filename, appname, **kw)

    def setUp(self):
        import tempfile
        self.tempfile = tempfile.mktemp()

    def tearDown(self):
        import os
        os.remove(self.tempfile)

    def test_no_conn_handler(self):
        factory = self._makeOne(self.tempfile, 'catalog')
        from zerodbext.catalog.catalog import Catalog
        catalog = factory()
        self.assertTrue(isinstance(catalog, Catalog))
        factory.db.close()

    def test_with_cache_size(self):
        factory = self._makeOne(self.tempfile, 'catalog', cache_size=1000)
        from zerodbext.catalog.catalog import Catalog
        catalog = factory()
        self.assertTrue(isinstance(catalog, Catalog))
        self.assertEqual(factory.db._cache_size, 1000)
        factory.db.close()

    def test_with_conn_handler(self):
        factory = self._makeOne(self.tempfile, 'catalog')
        from zerodbext.catalog.catalog import Catalog
        e = {}
        def handle(conn):
            e['conn'] = conn
        catalog = factory(handle)
        self.assertTrue(isinstance(catalog, Catalog))
        self.assertEqual(e['conn']._db, factory.db)
        factory.db.close()

class TestConnectionManager(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.catalog import ConnectionManager
        return ConnectionManager

    def _makeOne(self):
        klass = self._getTargetClass()
        return klass()

    def test_call(self):
        conn = DummyConnection()
        manager = self._makeOne()
        manager(conn)
        self.assertEqual(manager.conn, conn)

    def test_close(self):
        conn = DummyConnection()
        manager = self._makeOne()
        manager(conn)
        manager.close()
        self.assertEqual(conn.closed, True)

    def test_del(self):
        # Due to GC the conn may not be closed immediately on PyPy
        if platform.python_implementation() != "PyPy":
            conn = DummyConnection()
            manager = self._makeOne()
            manager(conn)
            del manager
            self.assertEqual(conn.closed, True)

    def test_commit(self):
        conn = DummyConnection()
        txn = DummyTransaction()
        manager = self._makeOne()
        manager(conn)
        manager.commit(txn)
        self.assertEqual(txn.committed, True)

class TestResultSetSize(unittest.TestCase):
    def _makeOne(self, i, total):
        from zerodbext.catalog.catalog import ResultSetSize
        return ResultSetSize(i, total)

    def test_ctor(self):
        size = self._makeOne(1, 2)
        self.assertEqual(size, 1)
        self.assertEqual(size.total, 2)

    def test_repr(self):
        size = self._makeOne(1, 2)
        self.assertEqual(repr(size), 'ResultSetSize(1, 2)')

    def test_pickles(self):
        import pickle
        size = self._makeOne(1, 2)
        unpickled = pickle.loads(pickle.dumps(size))
        self.assertEqual(unpickled, 1)
        self.assertEqual(unpickled.total, 2)
        self.assertEqual(repr(unpickled), 'ResultSetSize(1, 2)')

class DummyConnection:
    def close(self):
        self.closed = True

class DummyTransaction:
    def commit(self):
        self.committed = True

from zerodbext.catalog.catalog import ICatalogIndex
from zope.interface import implementer


@implementer(ICatalogIndex)
class DummyIndex(object):

    value = None
    docid = None
    limit = None
    sort_type = None

    def __init__(self, *arg, **kw):
        self.arg = arg
        self.kw = kw

    def index_doc(self, docid, value):
        self.docid = docid
        self.value = value
        return value

    def unindex_doc(self, docid):
        self.unindexed = docid

    def clear(self):
        self.cleared = True

    def reindex_doc(self, docid, object):
        self.reindexed_docid = docid
        self.reindexed_ob = object

    def apply(self, query):
        return self.arg[0]

    def apply_intersect(self, query, docids): # pragma: no cover
        if docids is None:
            return self.arg[0]
        L = []
        for docid in self.arg[0]:
            if docid in docids:
                L.append(docid)
        return L

    def sort(self, results, reverse=False, limit=None, sort_type=None):
        self.limit = limit
        self.sort_type = sort_type
        if reverse:
            return ['sorted3', 'sorted2', 'sorted1']
        return ['sorted1', 'sorted2', 'sorted3']



import bisect
import heapq
from itertools import islice

import six

from zope.interface import implementer

from zope.index.field import FieldIndex

from zerodbext.catalog.interfaces import ICatalogIndex
from zerodbext.catalog.indexes.common import CatalogIndex
from zerodbext.catalog import RangeValue

_marker = []

FWSCAN = 'fwscan'
NBEST = 'nbest'
TIMSORT = 'timsort'


@implementer(ICatalogIndex)
class CatalogFieldIndex(CatalogIndex, FieldIndex):
    """ Field indexing.

    Query types supported:

    - Eq

    - NotEq

    - Gt

    - Ge

    - Lt

    - Le

    - In

    - NotIn

    - Any

    - NotAny

    - InRange

    - NotInRange
    """

    def __init__(self, discriminator):
        if not callable(discriminator):
            if not isinstance(discriminator, six.string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        self._not_indexed = self.family.IF.Set()
        self.clear()

    def reindex_doc(self, docid, value):
        # the base index's index_doc method special-cases a reindex
        return self.index_doc(docid, value)

    def unindex_doc(self, docid):
        """See interface IInjection.

        Base class overridden to be able to unindex None values.
        """
        _not_indexed = self._not_indexed
        if docid in _not_indexed:
            _not_indexed.remove(docid)

        rev_index = self._rev_index
        value = rev_index.get(docid, _marker)
        if value is _marker:
            return # not in index

        del rev_index[docid]

        try:
            set = self._fwd_index[value]
            set.remove(docid)
        except KeyError:    #pragma NO COVERAGE
            # This is fishy, but we don't want to raise an error.
            # We should probably log something.
            # but keep it from throwing a dirty exception
            set = 1

        if not set:
            del self._fwd_index[value]

        self._num_docs.change(-1)

    def _indexed(self):
        return self._rev_index.keys()

    def sort(self, docids, reverse=False, limit=None, sort_type=None):
        if not docids:
            return []

        numdocs = self._num_docs.value
        if not numdocs:
            return []

        if limit is not None:
            limit = int(limit)
            if limit < 1:
                raise ValueError('limit must be 1 or greater')

        if reverse:
            return self.sort_reverse(docids, limit, numdocs, sort_type)
        else:
            return self.sort_forward(docids, limit, numdocs, sort_type)

    def sort_forward(self, docids, limit, numdocs, sort_type=None):

        rlen = len(docids)

        # See http://www.zope.org/Members/Caseman/ZCatalog_for_2.6.1
        # for an overview of why we bother doing all this work to
        # choose the right sort algorithm.

        if sort_type is None:

            if fwscan_wins(limit, rlen, numdocs):
                # forward scan beats both n-best and timsort reliably
                # if this is true
                sort_type = FWSCAN

            elif limit and nbest_ascending_wins(limit, rlen, numdocs):
                # nbest beats timsort reliably if this is true
                sort_type = NBEST

            else:
                sort_type = TIMSORT

        if sort_type == FWSCAN:
            return self.scan_forward(docids, limit)
        elif sort_type == NBEST:
            if limit is None:
                raise ValueError('nbest requires a limit')
            return self.nbest_ascending(docids, limit)
        elif sort_type == TIMSORT:
            return self.timsort_ascending(docids, limit)
        else:
            raise ValueError('Unknown sort type %s' % sort_type)

    def sort_reverse(self, docids, limit, numdocs, sort_type=None):
        if sort_type is None:
            # XXX this needs work.
            rlen = len(docids)
            if limit:
                if (limit < 300) or (limit/float(rlen) > 0.09):
                    sort_type = NBEST
                else:
                    sort_type = TIMSORT
            else:
                sort_type = TIMSORT

        if sort_type == NBEST:
            if limit is None:
                raise ValueError('nbest requires a limit')
            return self.nbest_descending(docids, limit)
        elif sort_type == TIMSORT:
            return self.timsort_descending(docids, limit)
        else:
            raise ValueError('Unknown sort type %s' % sort_type)

    def scan_forward(self, docids, limit=None):
        fwd_index = self._fwd_index

        n = 0
        for set in fwd_index.values():
            for docid in set:
                if docid in docids:
                    n+=1
                    yield docid
                    if limit and n >= limit:
                        return

    def nbest_ascending(self, docids, limit):
        if limit is None: #pragma NO COVERAGE
            raise RuntimeError('n-best used without limit')

        # lifted from heapq.nsmallest

        h = nsort(docids, self._rev_index)
        it = iter(h)
        result = sorted(islice(it, 0, limit))
        if not result: #pragma NO COVERAGE
            return
        insort = bisect.insort
        pop = result.pop
        los = result[-1]    # los --> Largest of the nsmallest
        for elem in it:
            if los <= elem:
                continue
            insort(result, elem)
            pop()
            los = result[-1]

        for value, docid in result:
            yield docid

    def nbest_descending(self, docids, limit):
        if limit is None: #pragma NO COVERAGE
            raise RuntimeError('N-Best used without limit')
        iterable = nsort(docids, self._rev_index)
        for value, docid in heapq.nlargest(limit, iterable):
            yield docid

    def timsort_ascending(self, docids, limit):
        return self._timsort(docids, limit, reverse=False)

    def timsort_descending(self, docids, limit):
        return self._timsort(docids, limit, reverse=True)

    def _timsort(self, docids, limit=None, reverse=False):
        n = 0
        marker = _marker
        _missing = []

        pairs = []
        for docid in docids:
            v = self._rev_index.get(docid, marker)
            if v is not marker:
                pairs.append((docid, v))

        for (docid, _) in sorted(pairs, key=lambda p: p[1], reverse=reverse):
            n += 1
            yield docid
            if limit and n >= limit:
                return

    def search(self, queries, operator='or'):
        sets = []
        for query in queries:
            if isinstance(query, RangeValue):
                query = query.as_tuple()
            else:
                query = (query, query)
            set = self.family.IF.multiunion(self._fwd_index.values(*query))
            sets.append(set)

        result = None

        if len(sets) == 1:
            result = sets[0]
        elif operator == 'and':
            sets.sort(key=len)
            for set in sets:
                result = self.family.IF.intersection(set, result)
        else:
            result = self.family.IF.multiunion(sets)

        return result

    def apply(self, query):
        if isinstance(query, dict):
            val = query['query']
            if isinstance(val, RangeValue):
                val = [val]
            elif not isinstance(val, (list, tuple)):
                val = [val]
            operator = query.get('operator', 'or')
            result = self.search(val, operator)
        else:
            if isinstance(query, tuple) and len(query) == 2:
                # b/w compat stupidity; this needs to die
                query = RangeValue(*query)
                query = [query]
            elif not isinstance(query, (list, tuple)):
                query = [query]
            result = self.search(query, 'or')

        return result

    def applyEq(self, value):
        return self.apply(value)

    def applyGe(self, min_value):
        return self.applyInRange(min_value, None)

    def applyLe(self, max_value):
        return self.applyInRange(None, max_value)

    def applyGt(self, min_value):
        return self.applyInRange(min_value, None, excludemin=True)

    def applyLt(self, max_value):
        return self.applyInRange(None, max_value, excludemax=True)

    def applyAny(self, values):
        queries = list(values)
        return self.search(queries, operator='or')

    def applyInRange(self, start, end, excludemin=False, excludemax=False):
        return self.family.IF.multiunion(
            self._fwd_index.values(
                start, end, excludemin=excludemin, excludemax=excludemax)
        )


def nsort(docids, rev_index):
    for docid in docids:
        try:
            yield (rev_index[docid], docid)
        except KeyError:
            continue


def fwscan_wins(limit, rlen, numdocs):
    """
    Primitive curve-fitting to see if forward scan will beat both
    nbest and timsort for a particular limit/rlen/numdocs tuple.  In
    sortbench tests up to 'numdocs' sizes of 65536, this curve fit had
    a 95%+ accuracy rate, except when 'numdocs' is < 64, then its
    lowest accuracy percentage was 83%.  Thus, it could still use some
    work, but accuracy at very small index sizes is not terribly
    important for the author.
    """
    docratio = rlen / float(numdocs)

    if limit:
        limitratio = limit / float(numdocs)
    else:
        limitratio = 1

    div = 65536.0

    if docratio >= 16384/div:
        # forward scan tends to beat nbest or timsort reliably when
        # the rlen is greater than a quarter of the number of
        # documents in the index
        return True

    if docratio >= 256/div:
        # depending on the limit ratio, forward scan still has a
        # chance to win over nbest or timsort even if the rlen is
        # smaller than a quarter of the number of documents in the
        # index, beginning reliably at a docratio of 512/65536.0.  XXX
        # It'd be nice to figure out a more concise way to express
        # this.
        if 512/div <= docratio < 1024/div and limitratio <= 4/div:
            return True
        elif  1024/div <= docratio < 2048/div and limitratio <= 32/div:
            return True
        elif 2048/div <= docratio < 4096/div and limitratio <= 128/div:
            return True
        elif 4096/div <= docratio < 8192/div and limitratio <= 512/div:
            return True
        elif 8192/div <= docratio < 16384/div and limitratio <= 4096/div:
            return True

    return False


def nbest_ascending_wins(limit, rlen, numdocs):
    """
    Primitive curve-fitting to see if nbest ascending will beat
    timsort for a particular limit/rlen/numdocs tuple.  XXX This needs
    work, particularly at small index sizes.  It is currently
    optimized for an index size of about 32768 (98% accuracy); it gets
    about 93% accuracy at index size 65536.
    """
    if not limit:
        # n-best can't be used without a limit
        return False
    limitratio = limit / float(numdocs)

    if numdocs <= 768:
        return True

    docratio = rlen / float(numdocs)
    div = 65536.0

    if docratio < 4096/div:
        # nbest tends to win when the rlen is less than about 6% of the
        # numdocs
        return True

    if docratio == 1 and limitratio <= 8192/div:
        return True
    elif 1 > docratio >= 32768/div and limitratio <= 4096/div:
        return True
    elif 32768/div > docratio >= 4096/div and limitratio <= 2048/div:
        return True

    return False



import six
from zope.interface import implementer
from persistent import Persistent

import BTrees

from BTrees.Length import Length

from zerodbext.catalog.interfaces import ICatalogIndex
from zerodbext.catalog.indexes.common import CatalogIndex

_marker = ()


@implementer(ICatalogIndex)
class CatalogPathIndex(CatalogIndex):

    """Index for model paths (tokens separated by '/' characters)

    A path index stores all path components of the physical path of an object.

    Internal datastructure:

    - a physical path of an object is split into its components

    - every component is kept as a  key of a OOBTree in self._indexes

    - the value is a mapping 'level of the path component' to
      'all docids with this path component on this level'


    Query types supported:

    - Eq

    - NotEq

    """
    useOperator = 'or'

    family = BTrees.family32

    def __init__(self, discriminator):
        if not callable(discriminator):
            if not isinstance(discriminator, six.string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        self._not_indexed = self.family.IF.Set()
        self.clear()

    def clear(self):
        self._depth = 0
        self._index = self.family.OO.BTree()
        self._unindex = self.family.IO.BTree()
        self._length = Length(0)

    def insertEntry(self, comp, id, level):
        """Insert an entry.

           comp is a path component
           id is the docid
           level is the level of the component inside the path
        """

        if comp not in self._index:
            self._index[comp] = self.family.IO.BTree()

        if level not in self._index[comp]:
            self._index[comp][level] = self.family.IF.TreeSet()

        self._index[comp][level].insert(id)
        if level > self._depth:
            self._depth = level

    def index_doc(self, docid, object):
        if callable(self.discriminator):
            value = self.discriminator(object, _marker)
        else:
            value = getattr(object, self.discriminator, _marker)

        if value is _marker:
            # unindex the previous value
            self.unindex_doc(docid)

            # Store docid in set of unindexed docids
            self._not_indexed.add(docid)

            return None

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object %s' %
                             value)

        if docid in self._not_indexed:
            # Remove from set of unindexed docs if it was in there.
            self._not_indexed.remove(docid)

        path = value

        if isinstance(path, (list, tuple)):
            path = '/'+ '/'.join(path[1:])

        comps = [c for c in path.split('/') if c]

        if docid not in self._unindex:
            self._length.change(1)

        for i in range(len(comps)):
            self.insertEntry(comps[i], docid, i)

        self._unindex[docid] = path
        return 1

    def unindex_doc(self, docid):
        _not_indexed = self._not_indexed
        if docid in _not_indexed:
            _not_indexed.remove(docid)

        if docid not in self._unindex:
            return

        comps =  self._unindex[docid].split('/')

        for level in range(len(comps[1:])):
            comp = comps[level+1]

            try:
                self._index[comp][level].remove(docid)

                if not self._index[comp][level]:
                    del self._index[comp][level]

                if not self._index[comp]:
                    del self._index[comp]
            except KeyError:
                pass

        self._length.change(-1)
        del self._unindex[docid]

    def _indexed(self):
        return self._unindex.keys()

    def search(self, path, default_level=0):
        """
        path is either a string representing a
        relative URL or a part of a relative URL or
        a tuple (path,level).

        level >= 0  starts searching at the given level
        level <  0  not implemented yet
        """
        if isinstance(path, six.string_types):
            level = default_level
        else:
            level = int(path[1])
            path  = path[0]

        comps = [c for c in path.split('/') if c]

        if len(comps) == 0:
            return self.family.IF.Set(self._unindex.keys())

        results = None
        if level >= 0:
            for i, comp in enumerate(comps):
                if comp not in self._index:
                    return self.family.IF.Set()
                if level+i not in self._index[comp]:
                    return self.family.IF.Set()
                results = self.family.IF.intersection(
                    results, self._index[comp][level+i])

        else:
            for level in range(self._depth + 1):
                ids = None
                for i, comp in enumerate(comps):
                    try:
                        ids = self.family.IF.intersection(
                            ids, self._index[comp][level+i])
                    except KeyError:
                        break
                else:
                    results = self.family.IF.union(results, ids)
        return results

    def numObjects(self):
        """ return the number distinct values """
        return len(self._unindex)

    def getEntryForObject(self, docid):
        """ Takes a document ID and returns all the information
            we have on that specific object.
        """
        return self._unindex.get(docid)

    def apply(self, query):
        """
        """
        level = 0
        operator = self.useOperator

        if isinstance(query, six.string_types):
            paths = [query]
        elif isinstance(query, (tuple, list)):
            paths = query
        else:
            paths = query.get('query', [])
            if isinstance(paths, six.string_types):
                paths = [ paths ]
            level = query.get('level', 0)
            operator = query.get('operator', self.useOperator).lower()

        sets = []
        for path in paths:
            sets.append(self.search(path, level))

        if operator == 'or':
            rs = self.family.IF.multiunion(sets)

        else:
            rs = None
            sets.sort(key=len)
            for set in sets:
                rs = self.family.IF.intersection(rs, set)
                if not rs:
                    break

        if rs:
            return rs
        else:
            return self.family.IF.Set()

    applyEq = apply



import six
from zope.interface import implementer

import BTrees

from zerodbext.catalog.interfaces import ICatalogIndex
from zerodbext.catalog.indexes.common import CatalogIndex

_marker = object()


@implementer(ICatalogIndex)
class CatalogPathIndex2(CatalogIndex):  #pragma NO COVERAGE
    """
    DEPRECATED

    Index for model paths (tokens separated by '/' characters or
    tuples representing a model path).

    A path index may be queried to obtain all subobjects (optionally
    limited by depth) of a certain path.

    This index differs from the original
    ``zerodbext.catalog.indexes.path.CatalogPath`` index inasmuch as it
    actually retains a graph representation of the objects in the path
    space instead of relying on 'level' information; query results
    relying on this level information may or may not be correct for
    any given tree.  Use of this index is suggested rather than the
    ``path`` index.

    Query types supported:

    Eq

    """
    attr_discriminator = None # b/w compat

    family = BTrees.family32

    def __init__(self, discriminator, attr_discriminator=None):
        if not callable(discriminator):
            if not isinstance(discriminator, six.string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        if attr_discriminator is not None and not callable(attr_discriminator):
            if not isinstance(attr_discriminator, six.string_types):
                raise ValueError('attr_discriminator value must be callable '
                                 'or a string')
        self.attr_discriminator = attr_discriminator
        self.clear()

    def clear(self):
        self.docid_to_path = self.family.IO.BTree()
        self.path_to_docid = self.family.OI.BTree()
        self.adjacency = self.family.IO.BTree()
        self.disjoint = self.family.OO.BTree()
        self.docid_to_attr = self.family.IO.BTree()

    def __len__(self):
        return len(self.docid_to_path)

    def __nonzero__(self):
        return True

    def _getPathTuple(self, path):
        if not path:
            raise ValueError('path must be nonempty (not %s)' % str(path))

        if isinstance(path, six.string_types):
            path = path.rstrip('/')
            path = tuple(path.split('/'))

        if path[0] != '':
            raise ValueError('Path must be absolute (not %s)' % str(path))

        return tuple(path)

    def _getObjectPath(self, object):
        if callable(self.discriminator):
            path = self.discriminator(object, _marker)
        else:
            path = getattr(object, self.discriminator, _marker)

        return path

    def _getObjectAttr(self, object):
        if callable(self.attr_discriminator):
            attr = self.attr_discriminator(object, _marker)
        else:
            attr = getattr(object, self.attr_discriminator, _marker)
        return attr

    def index_doc(self, docid, object):

        path = self._getObjectPath(object)

        if path is _marker:
            self.unindex_doc(docid)
            return None

        path = self._getPathTuple(path)

        if self.attr_discriminator is not None:
            attr = self._getObjectAttr(object)
            if attr is not _marker:
                self.docid_to_attr[docid] = attr

        self.docid_to_path[docid] = path
        self.path_to_docid[path] = docid

        if path in self.disjoint:
            self.adjacency[docid] = self.disjoint[path]
            del self.disjoint[path]

        if len(path) > 1:
            parent_path = tuple(path[:-1])
            parent_docid = self.path_to_docid.get(parent_path)
            if parent_docid is None:
                theset = self.disjoint.get(parent_path)
                if theset is None:
                    theset = self.family.IF.Set()
                    self.disjoint[parent_path] = theset
            else:
                theset = self.adjacency.get(parent_docid)
                if theset is None:
                    theset = self.family.IF.Set()
                    self.adjacency[parent_docid] = theset
            theset.insert(docid)

    def unindex_doc(self, docid):
        path = self.docid_to_path.get(docid)
        if path is None:
            return

        if len(path) > 1:
            parent_path = tuple(path[:-1])
            parent_docid = self.path_to_docid.get(parent_path)
            if parent_docid is not None:  # might be disjoint
                self.adjacency[parent_docid].remove(docid)
                if not self.adjacency[parent_docid]:
                    del self.adjacency[parent_docid]
            else:
                self.disjoint[parent_path].remove(docid)
                if not self.disjoint[parent_path]:
                    del self.disjoint[parent_path]

        stack = [docid]

        while stack:
            docid = stack.pop()
            path = self.docid_to_path[docid]
            del self.path_to_docid[path]
            del self.docid_to_path[docid]
            if docid in self.docid_to_attr:
                del self.docid_to_attr[docid]
            next_docids = self.adjacency.get(docid)
            if next_docids is None:
                next_docids = self.disjoint.get(path)
                if next_docids is not None:
                    del self.disjoint[path]
                    stack.extend(next_docids)
            else:
                del self.adjacency[docid]
                stack.extend(next_docids)

    def reindex_doc(self, docid, object):
        path = self._getPathTuple(self._getObjectPath(object))

        if self.docid_to_path.get(docid) != path:
            self.unindex_doc(docid)
            self.index_doc(docid, object)
            return True

        else:
            if self.attr_discriminator is not None:
                attr = self._getObjectAttr(object)
                if docid in self.docid_to_attr:
                    if attr is _marker:
                        del self.docid_to_attr[docid]
                        return True
                    elif attr != self.docid_to_attr[docid]:
                        self.docid_to_attr[docid] = attr
                        return True
                else:
                    if attr is not _marker:
                        self.docid_to_attr[docid] = attr
                        return True
            return False

    def _indexed(self):
        return self.docid_to_path.keys()

    def search(self, path, depth=None, include_path=False, attr_checker=None):
        """ Provided a path string (e.g. ``/path/to/object``) or a
        path tuple (e.g. ``('', 'path', 'to', 'object')``, or a path
        list (e.g. ``['', 'path', 'to' object'])``), search the index
        for document ids representing subelements of the path
        specified by the path argument.

        If the ``path`` argment is specified as a tuple or list, its
        first element must be the empty string.  If the ``path``
        argument is specified as a string, it must begin with a ``/``
        character.  In other words, paths passed to the ``search``
        method must be absolute.

        If the ``depth`` argument is specified, return only documents
        at this depth and below.  Depth ``0`` will returns the empty
        set (or only the docid for the ``path`` specified if
        ``include_path`` is also True).  Depth ``1`` will return
        docids related to direct subobjects of the path (plus the
        docid for the ``path`` specified if ``include_path`` is also
        True).  Depth ``2`` will return docids related to direct
        subobjects and the docids of the children of those subobjects,
        and so on.

        If ``include_path`` is False, the docid of the object
        specified by the ``path`` argument is *not* returned as part
        of the search results.  If ``include_path`` is True, the
        object specified by the ``path`` argument *is* returned as
        part of the search results.

        If ``attr_checker`` is not None, it must be a callback that
        accepts two arguments: the first argument will be the
        attribute value found, the second argument is a sequence of
        all previous attributes encountered during this search (in
        path order).  If ``attr_checker`` returns True, traversal will
        continue; otherwise, traversal will cease.
        """
        if attr_checker is None:
            return self._simple_search(path, depth, include_path)
        else:
            return self._attr_search(path, depth, include_path, attr_checker)

    def _simple_search(self, path, depth, include_path):
        """ Codepath taken when no attr checker is used """
        path = self._getPathTuple(path)
        sets = []

        if include_path:
            try:
                docid = self.path_to_docid[path]
            except KeyError:
                pass # XXX should we just return an empty set?
            else:
                sets.append(self.family.IF.Set([docid]))

        stack = [path]
        plen = len(path)

        while stack:
            nextpath = stack.pop()
            if depth is not None and len(nextpath) - plen >= depth:
                continue
            try:
                docid = self.path_to_docid[nextpath]
            except KeyError:
                continue    # XXX we can't search from an unindexed root path?
            try:
                theset = self.adjacency[docid]
            except KeyError:
                pass
            else:
                sets.append(theset)
                for docid in theset:
                    try:
                        newpath = self.docid_to_path[docid]
                    except KeyError:
                        continue
                    stack.append(newpath)

        return self.family.IF.multiunion(sets)

    def _attr_search(self, path, depth, include_path, attr_checker):
        """ Codepath taken when an attr checker is used """
        path = self._getPathTuple(path)

        leading_attrs = []
        result = {}
        plen = len(path)

        # make sure we get "leading" attrs
        for p in range(plen-1):
            subpath = path[:p+1]
            try:
                docid = self.path_to_docid[subpath]
            except KeyError:
                continue  # XXX should we just return an empty set?
            attr = self.docid_to_attr.get(docid, _marker)
            if attr is not _marker:
                remove_from_closest(result, subpath, docid)
                leading_attrs.append(attr)
                result[subpath] = ((docid, leading_attrs[:]),
                                   self.family.IF.Set())

        stack = [(path, leading_attrs)]
        attrset = self.family.IF.Set()

        while stack:
            nextpath, attrs = stack.pop()
            try:
                docid = self.path_to_docid[nextpath]
            except KeyError:
                continue # XXX we can't search from an unindexed root path?
            attr = self.docid_to_attr.get(docid, _marker)
            if attr is _marker:
                if include_path and nextpath == path:
                    add_to_closest(
                        result, nextpath, self.family.IF.Set([docid]))
                if depth is not None and len(nextpath) - plen >= depth:
                    continue
            else:
                remove_from_closest(result, nextpath, docid)
                attrs.append(attr)
                if nextpath == path:
                    if include_path:
                        attrset = self.family.IF.Set([docid])
                    else:
                        attrset = self.family.IF.Set()
                else:
                    attrset = self.family.IF.Set([docid])
                result[nextpath] = ((docid, attrs), attrset)
                if depth is not None and len(nextpath) - plen >= depth:
                    continue
            try:
                theset = self.adjacency[docid]
            except KeyError:
                pass
            else:
                add_to_closest(result, nextpath, theset)
                for docid in theset:
                    try:
                        newpath = self.docid_to_path[docid]
                    except KeyError:
                        continue
                    stack.append((newpath, attrs[:]))

        return attr_checker(result.values())

    def apply_intersect(self, query, docids):
        """ Default apply_intersect implementation """
        result = self.apply(query)
        if docids is None:
            return result
        return self.family.IF.weightedIntersection(result, docids)[1]

    def apply(self, query):
        """ Search the path index using the query.  If ``query`` is a
        string, a tuple, or a list, it is treated as the ``path``
        argument to use to search.  If it is any other object, it is
        assumed to be a dictionary with at least a value for the
        ``query`` key, which is treated as a path.  The dictionary can
        also optionally specify the ``depth`` and whether to include
        the docid referenced by the path argument (the ``query`` key)
        in the set of docids returned (``include_path``).  See the
        documentation for the ``search`` method of this class to
        understand paths, depths, and the ``include_path`` argument.
        """
        if isinstance(query, tuple(list(six.string_types) + [tuple, list])):
            path = query
            depth = None
            include_path = False
            attr_checker = None
        else:
            path = query['query']
            depth = query.get('depth', None)
            include_path = query.get('include_path', False)
            attr_checker = query.get('attr_checker', None)

        return self.search(path, depth, include_path, attr_checker)

    def applyEq(self, query):
        return self.apply(query)

def add_to_closest(sofar, thispath, theset):
    paths = sorted(sofar.keys(), reverse=True)
    for path in paths:
        pathlen = len(path)
        if thispath[:pathlen] == path:
            sofar[path][1].update(theset)
            break

def remove_from_closest(sofar, thispath, docid):
    paths = sorted(sofar.keys(), reverse=True)
    for path in paths:
        pathlen = len(path)
        if thispath[:pathlen] == path:
            theset = sofar[path][1]
            if docid in theset:
                theset.remove(docid)
            break




import unittest
from zerodbext.catalog import query as q


class TestQueryBase(object):
    def test_it(self):
        from zerodbext.catalog.catalog import Catalog
        from zerodbext.catalog.indexes.field import CatalogFieldIndex
        from zerodbext.catalog.indexes.keyword import CatalogKeywordIndex
        from zerodbext.catalog.indexes.text import CatalogTextIndex

        catalog = Catalog()
        catalog['name'] = CatalogFieldIndex('name')
        catalog['title'] = CatalogFieldIndex('title')
        catalog['text'] = CatalogTextIndex('text')
        catalog['allowed'] = CatalogKeywordIndex('allowed')

        catalog.index_doc(1, Content('name1', 'title1', 'body one', ['a']))
        catalog.index_doc(2, Content('name2', 'title2', 'body two', ['b']))
        catalog.index_doc(3, Content('name3', 'title3', 'body three', ['c']))
        catalog.index_doc(4, Content('name4', None, 'body four',['a', 'b']))
        catalog.index_doc(5, Content('name5', 'title5', 'body five',
                                     ['a', 'b', 'c']))
        catalog.index_doc(6, Content('name6', 'title6', 'body six',['d']))

        numdocs, result = catalog.query(
            self.query, sort_index='name', limit=5, names=dict(body='body'))
        self.assertEqual(numdocs, 2)
        self.assertEqual(list(result), [4, 5])


class TestQueryWithCQE(unittest.TestCase, TestQueryBase):
    query = (
        "(allowed == 'a' and allowed == 'b' and "
        "(name == 'name1' or name == 'name2' or name == 'name3' or "
        "name == 'name4' or name == 'name5') and not(title == 'title3')) and "
        "body in text"
        )


class TestQueryWithPythonQueryObjects(unittest.TestCase, TestQueryBase):
    query = (
        q.All('allowed', ['a', 'b']) &
        q.Any('name', ['name1', 'name2', 'name3', 'name4', 'name5']) &
        q.Not(q.Eq('title', 'title3')) &
        q.Contains('text', 'body')
        )


try:
    import ast
except ImportError: #pragma NO COVERAGE
    del TestQueryWithCQE
    del TestQueryWithPythonQueryObjects

class Content(object):
    def __init__(self, name, title, text, allowed):
        self.name = name
        if title:
            self.title = title
        self.text = text
        self.allowed = allowed





import six
import BTrees
import sys

try:
    import ast
    ast_support = True
except ImportError:  # pragma NO COVERAGE
    ast_support = False


_marker = object()


class Query(object):
    """
    Base class for all elements that make up queries.
    """
    __parent__ = None
    __name__ = None

    def __and__(self, right):
        self._check_type("and", right)
        return And(self, right)

    def __or__(self, right):
        self._check_type("or", right)
        return Or(self, right)

    def _check_type(self, setop, operand):
        if not isinstance(operand, Query):
            raise TypeError(
                "TypeError: unsupported operand types for %s: %s %s" %
                (setop, type(self), type(operand)))

    def iter_children(self):
        return ()

    def print_tree(self, out=sys.stdout, level=0):
        six.print_('  ' * level + str(self), file=out)
        for child in self.iter_children():
            child.print_tree(out, level + 1)

    def _optimize(self):
        """
        If subtree represented by this node can be transformed into a more
        optimal subtree, return the transformed subtree, otherwise return self.
        """
        return self


class Comparator(Query):
    """
    Base class for all comparators used in queries.
    """

    def __init__(self, index_name, value):
        self.index_name = index_name
        self._value = value

    def _get_index(self, catalog):
        return catalog[self.index_name]

    def _get_value(self, names, value=_marker):
        if value is _marker:
            value = self._value
        if isinstance(value, list):
            return [self._get_value(names, child) for child in value]
        elif isinstance(value, tuple):
            return tuple(self._get_value(names, child) for child in value)
        elif isinstance(value, Name):
            name = value.name
            if not names or name not in names:
                raise NameError("No value passed in for name: %s" % name)
            return names[name]
        return value

    def __str__(self):
        return ' '.join((self.index_name, self.operator, repr(self._value)))

    def __eq__(self, other):
        return (self.index_name == other.index_name and
                self._value == other._value)


class Contains(Comparator):
    """Contains query.

    CQE equivalent: 'foo' in index
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyContains(self._get_value(names))

    def __str__(self):
        return '%s in %s' % (repr(self._value), self.index_name)

    def negate(self):
        return DoesNotContain(self.index_name, self._value)


class DoesNotContain(Comparator):
    """CQE equivalent: 'foo' not in index
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyDoesNotContain(self._get_value(names))

    def __str__(self):
        return '%s not in %s' % (repr(self._value), self.index_name)

    def negate(self):
        return Contains(self.index_name, self._value)


class Eq(Comparator):
    """Equals query.

    CQE equivalent:  index == 'foo'
    """
    operator = '=='

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyEq(self._get_value(names))

    def negate(self):
        return NotEq(self.index_name, self._value)


class NotEq(Comparator):
    """Not equal query.

    CQE eqivalent: index != 'foo'
    """
    operator = '!='

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyNotEq(self._get_value(names))

    def negate(self):
        return Eq(self.index_name, self._value)


class Gt(Comparator):
    """ Greater than query.

    CQE equivalent: index > 'foo'
    """
    operator = '>'

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyGt(self._get_value(names))

    def negate(self):
        return Le(self.index_name, self._value)


class Lt(Comparator):
    """ Less than query.

    CQE equivalent: index < 'foo'
    """
    operator = '<'

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyLt(self._get_value(names))

    def negate(self):
        return Ge(self.index_name, self._value)


class Ge(Comparator):
    """Greater (or equal) query.

    CQE equivalent: index >= 'foo'
    """
    operator = '>='

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyGe(self._get_value(names))

    def negate(self):
        return Lt(self.index_name, self._value)


class Le(Comparator):
    """Less (or equal) query.

    CQE equivalent: index <= 'foo
    """
    operator = '<='

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyLe(self._get_value(names))

    def negate(self):
        return Gt(self.index_name, self._value)


class Any(Comparator):
    """Any of query.

    CQE equivalent: index in any(['foo', 'bar'])
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyAny(self._get_value(names))

    def negate(self):
        return NotAny(self.index_name, self._value)

    def __str__(self):
        return '%s in any(%s)' % (self.index_name, repr(self._value))


class NotAny(Comparator):
    """Not any of query (ie, None of query)

    CQE equivalent: index not in any(['foo', 'bar'])
    """
    operator = 'not any'

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyNotAny(self._get_value(names))

    def negate(self):
        return Any(self.index_name, self._value)

    def __str__(self):
        return '%s not in any(%s)' % (self.index_name, repr(self._value))


class All(Comparator):
    """All query.

    CQE equivalent: index in all(['foo', 'bar'])
    """
    operator = 'all'

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyAll(self._get_value(names))

    def negate(self):
        return NotAll(self.index_name, self._value)

    def __str__(self):
        return '%s in all(%s)' % (self.index_name, repr(self._value))


class NotAll(Comparator):
    """NotAll query.

    CQE equivalent: index not in all(['foo', 'bar'])
    """
    operator = 'not all'

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyAll(self._get_value(names))

    def negate(self):
        return All(self.index_name, self._value)

    def __str__(self):
        return '%s not in all(%s)' % (self.index_name, repr(self._value))


class _Range(Comparator):

    @classmethod
    def fromGTLT(cls, start, end):
        assert isinstance(start, (Gt, Ge))
        if isinstance(start, Gt):
            start_exclusive = True
        else:
            start_exclusive = False

        assert isinstance(end, (Lt, Le))
        if isinstance(end, Lt):
            end_exclusive = True
        else:
            end_exclusive = False

        assert start.index_name == end.index_name
        return cls(start.index_name, start._value, end._value,
                   start_exclusive, end_exclusive)

    def __init__(self, index_name, start, end,
                 start_exclusive=False, end_exclusive=False):
        self.index_name = index_name
        self._start = start
        self._end = end
        self.start_exclusive = start_exclusive
        self.end_exclusive = end_exclusive

    def _get_start(self, names):
        value = self._start
        if isinstance(value, Name):
            name = value.name
            if name not in names:
                raise NameError("No value passed in for name: %s" % name)
            return names[name]
        return value

    def _get_end(self, names):
        value = self._end
        if isinstance(value, Name):
            name = value.name
            if name not in names:
                raise NameError("No value passed in for name: %s" % name)
            return names[name]
        return value

    def __str__(self):
        s = [repr(self._start)]
        if self.start_exclusive:
            s.append('<')
        else:
            s.append('<=')
        s.append(self.index_name)
        if self.end_exclusive:
            s.append('<')
        else:
            s.append('<=')
        s.append(repr(self._end))
        return ' '.join(s)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.index_name == other.index_name and
                self._start == other._start and
                self._end == other._end and
                self.start_exclusive == other.start_exclusive and
                self.end_exclusive == other.end_exclusive)


class InRange(_Range):
    """ Index value falls within a range.

    CQE eqivalent: lower < index < upper
                   lower <= index <= upper
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyInRange(
            self._get_start(names), self._get_end(names),
            self.start_exclusive, self.end_exclusive)

    def negate(self):
        return NotInRange(self.index_name, self._start, self._end,
                          self.start_exclusive, self.end_exclusive)


class NotInRange(_Range):
    """ Index value falls outside a range.

    CQE eqivalent: not(lower < index < upper)
                   not(lower <= index <= upper)
    """

    def _apply(self, catalog, names):
        index = self._get_index(catalog)
        return index.applyNotInRange(
            self._get_start(names), self._get_end(names),
            self.start_exclusive, self.end_exclusive)

    def __str__(self):
        return 'not(%s)' % _Range.__str__(self)

    def negate(self):
        return InRange(self.index_name, self._start, self._end,
                       self.start_exclusive, self.end_exclusive)


class BoolOp(Query):
    """
    Base class for Or and And operators.
    """
    family = BTrees.family32

    def __init__(self, *queries):
        arguments = []
        for query in queries:
            # If argument is of the same type, can promote its arguments up
            # to here.
            if type(query) == type(self):
                arguments.extend(query.queries)
            else:
                arguments.append(query)
        self.queries = arguments

    def __str__(self):
        return type(self).__name__

    def iter_children(self):
        for query in self.queries:
            yield query

    def _optimize(self):
        self.queries = [query._optimize() for query in self.queries]
        new_me = self._optimize_eq()
        if new_me is not None:
            return new_me
        new_me = self._optimize_not_eq()
        if new_me is not None:
            return new_me
        return self

    def _optimize_eq(self):
        # If all queries are Eq operators for the same index, we can replace
        # this And or Or with an All or Any node.
        queries = list(self.queries)
        query = queries.pop(0)
        if type(query) != Eq:
            return None
        index_name = query.index_name
        values = [query._value]
        while queries:
            query = queries.pop(0)
            if type(query) != Eq or query.index_name != index_name:
                return None
            values.append(query._value)

        # All queries are Eq operators for the same index.
        if type(self) == Or:
            return Any(index_name, values)
        return All(index_name, values)

    def _optimize_not_eq(self):
        # If all queries are NotEq operators for the same index, we can
        # replace this And or Or with a NotAll or NotAny node.
        queries = list(self.queries)
        query = queries.pop(0)
        if type(query) != NotEq:
            return None
        index_name = query.index_name
        values = [query._value]
        while queries:
            query = queries.pop(0)
            if type(query) != NotEq or query.index_name != index_name:
                return None
            values.append(query._value)

        # All queries are Eq operators for the same index.
        if type(self) == Or:
            return NotAll(index_name, values)
        return NotAny(index_name, values)


class Or(BoolOp):
    """Boolean Or of multiple queries."""

    def _apply(self, catalog, names):
        # XXX Try to figure out when we need weightedOr and when we can
        # just use union or multiunion.
        queries = self.queries
        result = queries[0]._apply(catalog, names)
        for query in queries[1:]:
            next_result = query._apply(catalog, names)
            if len(result) == 0:
                result = next_result
            elif len(next_result) > 0:
                _, result = self.family.IF.weightedUnion(result, next_result)
        return result

    def negate(self):
        neg_queries = [query.negate() for query in self.queries]
        return And(*neg_queries)

    def _optimize(self):
        new_self = BoolOp._optimize(self)
        if self is not new_self:
            return new_self

        # There might be a combination of Gt/Ge and Lt/Le operators for the
        # same index that could be used to compose a NotInRange.
        uppers = {}
        lowers = {}
        queries = list(self.queries)

        def process_range(i_lower, query_lower, i_upper, query_upper):
            queries[i_lower] = NotInRange.fromGTLT(
                query_lower.negate(), query_upper.negate())
            queries[i_upper] = None

        for i in six.moves.xrange(len(queries)):
            query = queries[i]
            if type(query) in (Lt, Le):
                match = uppers.get(query.index_name)
                if match is not None:
                    i_upper, query_upper = match
                    process_range(i, query, i_upper, query_upper)
                else:
                    lowers[query.index_name] = (i, query)

            elif type(query) in (Gt, Ge):
                match = lowers.get(query.index_name)
                if match is not None:
                    i_lower, query_lower = match
                    process_range(i_lower, query_lower, i, query)
                else:
                    uppers[query.index_name] = (i, query)

        queries = [q for q in queries if q]
        if len(queries) == 1:
            return queries[0]

        self.queries = queries
        return self


class And(BoolOp):
    """Boolean And of multiple queries."""

    def _apply(self, catalog, names):
        # XXX Try to figure out when we need weightedIntersection and when we
        # can just use intersection.
        IF = self.family.IF
        queries = self.queries
        result = queries[0]._apply(catalog, names)
        for query in queries[1:]:
            if len(result) == 0:
                return IF.Set()
            next_result = query._apply(catalog, names)
            if len(next_result) == 0:
                return IF.Set()
            _, result = IF.weightedIntersection(result, next_result)
        return result

    def negate(self):
        neg_queries = [query.negate() for query in self.queries]
        return Or(*neg_queries)

    def _optimize(self):
        new_self = BoolOp._optimize(self)
        if self is not new_self:
            return new_self

        # There might be a combination of Gt/Ge and Lt/Le operators for the
        # same index that could be used to compose an InRange.
        uppers = {}
        lowers = {}
        queries = list(self.queries)

        def process_range(i_lower, query_lower, i_upper, query_upper):
            queries[i_lower] = InRange.fromGTLT(query_lower, query_upper)
            queries[i_upper] = None

        for i in six.moves.xrange(len(queries)):
            query = queries[i]
            if type(query) in (Gt, Ge):
                match = uppers.get(query.index_name)
                if match is not None:
                    i_upper, query_upper = match
                    process_range(i, query, i_upper, query_upper)
                else:
                    lowers[query.index_name] = (i, query)

            elif type(query) in (Lt, Le):
                match = lowers.get(query.index_name)
                if match is not None:
                    i_lower, query_lower = match
                    process_range(i_lower, query_lower, i, query)
                else:
                    uppers[query.index_name] = (i, query)

        queries = [q for q in queries if q]
        if len(queries) == 1:
            return queries[0]

        self.queries = queries
        return self


class Not(Query):
    """Negation of a query."""

    def __init__(self, query):
        self.query = query

    def __str__(self):
        return 'Not'

    def iter_children(self):
        yield self.query

    def negate(self):
        return self.query

    def _apply(self, catalog, names):
        return self.query.negate()._apply(catalog, names)

    def _optimize(self):
        return self.query.negate()._optimize()


class Name(object):
    """
    A variable name in an expression, evaluated at query time.  Can be used
    to defer evaluation of variables used inside of expressions until query
    time.

    Example::

        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import Name

        # Define query at module scope
        find_cats = Eq('color', Name('color')) & Eq('sex', Name('sex'))

        # Use query in a search function, evaluating color and sex at the
        # time of the query
        def search_cats(catalog, resolver, color='tabby', sex='female'):
            # Let resolver be some function which can retrieve a cat object
            # from your application given a docid.
            params = dict(color=color, sex=sex)
            count, docids = catalog.query(find_cats, params)
            for docid in docids:
                yield resolver(docid)
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.name)

    __str__ = __repr__

    def __eq__(self, right):
        if isinstance(right, Name):
            return right.name == self.name
        return False


class _AstParser(object):
    """
    Uses Python's ast module to parse an expression into an abstract syntax
    tree.  It then walks the tree and constructs a tree of Query objects.  Take
    the following query:

        >>> expr = "a == 1 or (b == 2 or b == 3)"

    The ast for this expression looks like this:

        >>> _print_ast(expr)
        <_ast.Module object at 0xb7377d4c>
          <_ast.Expr object at 0xb737e44c>
            <_ast.BoolOp object at 0x88f13cc>
              <_ast.And object at 0x88ec0ac>
              <_ast.Compare object at 0x88f13ac>
                <_ast.Name object at 0x88f13ec>
                  <_ast.Load object at 0x88ea86c>
                <_ast.Eq object at 0x88ece2c>
                <_ast.Num object at 0x88f14cc>
              <_ast.BoolOp object at 0x88f14ec>
                <_ast.Or object at 0x88ec14c>
                <_ast.Compare object at 0x88f154c>
                  <_ast.Name object at 0x88f15ac>
                    <_ast.Load object at 0x88ea86c>
                  <_ast.Eq object at 0x88ece2c>
                  <_ast.Num object at 0x88f15cc>
                <_ast.Compare object at 0x88f162c>
                  <_ast.Name object at 0x88f168c>
                    <_ast.Load object at 0x88ea86c>
                  <_ast.Eq object at 0x88ece2c>
                  <_ast.Num object at 0x88f16ac>

    _ast.Module is always the root of any tree returned by the ast parser. It
    is a requirement for _AstParser that the _ast.Module node contain only a
    single child node of type _ast.Expr, which represents the expression we
    are trying to transform into a query. The _ast.Expr node will always only
    have a single child which is the root of the expression tree,
    _ast.BoolOp in the above example.

    The walk method is the driver for constructing the query tree.  It performs
    a depth first traversal of the ast.  For each node in the ast it checks to
    see if we have a method for processing that node type.  Node processors are
    all named 'process_NodeType' where NodeType is the name of the class of the
    ast node, ie type(node).__name__.  Each processor method is passed the
    current node and its children which have already been processed.  In this
    way the query tree is built from the ast from the bottom up.
    """

    def __init__(self, expr):
        self.expr = expr

    def parse(self):
        statements = ast.parse(self.expr).body
        if len(statements) > 1:
            raise ValueError("Can only process single expression.")
        expr_tree = statements[0]
        if not isinstance(expr_tree, ast.Expr):
            raise ValueError("Not an expression.")

        result = self.walk(expr_tree.value)
        return result

    def walk(self, tree):

        def visit(node):
            children = [visit(child) for child in ast.iter_child_nodes(node)]
            name = 'process_%s' % type(node).__name__
            processor = getattr(self, name, None)
            if processor is None:
                raise ValueError(
                    "Unable to parse expression.  Unhandled expression "
                    "element: %s" % type(node).__name__)
            return processor(node, children)

        return visit(tree)

    def process_Load(self, node, children):
        pass

    def process_Name(self, node, children):
        return node

    def process_Attribute(self, node, children):
        name = children[0]
        dotted_name = ast.Name()
        dotted_name.id = '.'.join((name.id, node.attr))
        return dotted_name

    def process_Str(self, node, children):
        return node.s

    def process_Num(self, node, children):
        return node.n

    def process_List(self, node, children):
        l = list(children[:-1])
        for i in six.moves.xrange(len(l)):
            if isinstance(l[i], ast.Name):
                l[i] = self._value(l[i])
        return l

    def process_Tuple(self, node, children):
        return tuple(self.process_List(node, children))

    def process_Eq(self, node, children):
        return self.process_comparator(Eq)

    def process_NotEq(self, node, children):
        return self.process_comparator(NotEq)

    def process_Lt(self, node, children):
        return self.process_comparator(Lt)

    def process_LtE(self, node, children):
        return self.process_comparator(Le)

    def process_Gt(self, node, children):
        return self.process_comparator(Gt)

    def process_GtE(self, node, children):
        return self.process_comparator(Ge)

    def process_comparator(self, cls):

        def factory(left, right):
            return cls(self._index_name(left), self._value(right))

        factory.type = cls
        return factory

    def process_In(self, node, children):

        def factory(left, right):
            if callable(right):  # any or all, see process_Call
                return right(self._index_name(left))
            return Contains(self._index_name(right), self._value(left))

        factory.type = Contains
        return factory

    def process_NotIn(self, node, children):

        def factory(left, right):
            if callable(right):  # any or all, see process_Call
                return right(self._index_name(left)).negate()
            return DoesNotContain(self._index_name(right), self._value(left))

        factory.type = DoesNotContain
        return factory

    def process_Not(self, node, children):
        return Not

    def process_UnaryOp(self, node, children):
        operator, query = children
        return operator(query)

    def process_USub(self, node, children):
        return Not

    def process_Compare(self, node, children):
        # Python allows arbitrary chaining of comparisons, ie:
        #   x == y == z != abc
        #   x < y >= z
        #
        # For our purposes, though, we are only interested in two basic forms:
        #   index_name <comparison_operator> value
        # or
        #   start [<|<=] index_name [<|<=] end
        #
        # Where the second form maps to an InRange comparator and the first
        # form matches any of the other comparators.  Arbitrary chaining as
        # shown above is not supported.
        if len(children) == 3:
            # Simple binary form
            left, factory, right = children
            return factory(left, right)
        elif len(children) == 5:
            # Range expression
            start, f1, f2, index_name, end = children
            op1, op2 = f1.type, f2.type
            if op1 in (Lt, Le) and op2 in (Lt, Le):
                if op1 is Lt:
                    start_exclusive = True
                else:
                    start_exclusive = False
                if op2 is Lt:
                    end_exclusive = True
                else:
                    end_exclusive = False
                return InRange(self._index_name(index_name),
                               self._value(start),
                               self._value(end),
                               start_exclusive,
                               end_exclusive)
        raise ValueError(
            "Bad expression: unsupported chaining of comparators.")

    def process_BitOr(self, node, children):
        return Or

    def process_BitAnd(self, node, children):
        return And

    def process_BinOp(self, node, children):
        left, operator, right = children
        if not isinstance(left, Query):
            raise ValueError(
                "Bad expression: left operand for %s must be a result set." %
                operator.__name__)
        if not isinstance(right, Query):
            raise ValueError(
                "Bad expression: right operand for %s must be a result set." %
                operator.__name__)
        return operator(left, right)

    def process_Or(self, node, children):
        return Or

    def process_And(self, node, children):
        return And

    def process_BoolOp(self, node, children):
        operator = children.pop(0)
        for child in children:
            if not isinstance(child, Query):
                raise ValueError(
                    "Bad expression: All operands for %s must be result sets."
                    % operator.__name__)

        return operator(*children)

    def process_Call(self, node, children):
        func = children.pop(0)
        name = getattr(func, 'id', str(node.func))
        if name not in ('any', 'all'):
            raise ValueError(
                "Bad expression: Illegal function call in expression: %s" %
                name)
        if len(children) != 1:
            raise ValueError(
                "Bad expression: Wrong number of arguments to %s" % name)

        values = children[0]
        if name == 'any':
            comparator = Any
        else:
            comparator = All

        def factory(index_name):
            return comparator(index_name, self._value(values))

        return factory

    def _index_name(self, node):
        if not isinstance(node, ast.Name):
            raise ValueError("Index name must be a name.")
        return node.id

    def _value(self, node):
        if isinstance(node, ast.Name):
            return Name(node.id)
        return node


def optimize(query):
    if isinstance(query, Query):
        return query._optimize()
    return query


def parse_query(expr, optimize_query=True):
    """
    Parses the given expression string and returns a query object.  Requires
    Python >= 2.6.
    """
    if not ast_support:
        raise NotImplementedError("Parsing of CQEs requires Python >= 2.6")
    query = _AstParser(expr).parse()
    if optimize_query:
        query = optimize(query)
    return query


def _print_ast(expr):  # pragma NO COVERAGE
    """
    Useful method for visualizing AST trees while debugging.
    """
    tree = ast.parse(expr)

    def visit(node, level):
        print('  ' * level + str(node))
        for child in ast.iter_child_nodes(node):
            visit(child, level + 1)
    visit(tree, 0)



import unittest
from zerodbext.catalog.query import ast_support


class ComparatorTestBase(unittest.TestCase):

    def _makeOne(self, index_name, value):
        return self._getTargetClass()(index_name, value)


class TestQuery(unittest.TestCase):

    def _makeOne(self):
        from zerodbext.catalog.query import Query as cls
        return cls()

    def test_and(self):
        from zerodbext.catalog.query import And
        a = self._makeOne()
        b = self._makeOne()
        result = a & b
        self.assertTrue(isinstance(result, And))
        self.assertEqual(result.queries[0], a)
        self.assertEqual(result.queries[1], b)

    def test_and_type_error(self):
        a = self._makeOne()
        self.assertRaises(TypeError, a.__and__, 2)

    def test_or(self):
        from zerodbext.catalog.query import Or
        a = self._makeOne()
        b = self._makeOne()
        result = a | b
        self.assertTrue(isinstance(result, Or))
        self.assertEqual(result.queries[0], a)
        self.assertEqual(result.queries[1], b)

    def test_or_type_error(self):
        a = self._makeOne()
        self.assertRaises(TypeError, a.__or__, 2)

    def test_iter_children(self):
        a = self._makeOne()
        self.assertEqual(a.iter_children(), ())

    def test_print_tree(self):
        from zerodbext.catalog.query import Query

        class Derived(Query):

            def __init__(self, name):
                self.name = name
                self.children = []

            def __str__(self):
                return self.name

            def iter_children(self):
                return self.children

        from six import StringIO
        a = Derived('A')
        b = Derived('B')
        c = Derived('C')
        a.children.append(b)
        a.children.append(c)

        buf = StringIO()
        a.print_tree(buf)
        self.assertEqual(buf.getvalue(), 'A\n  B\n  C\n')


class TestComparator(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Comparator
        return Comparator

    def test_ctor(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.index_name, 'index')
        self.assertEqual(inst._value, 'val')

    def test_eq(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst, self._makeOne('index', 'val'))


class TestContains(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Contains
        return Contains

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.contains, 'val')

    def test_apply_w_name(self):
        from zerodbext.catalog.query import Name
        catalog = DummyCatalog()
        inst = self._makeOne('index', Name('foo'))
        result = inst._apply(catalog, {'foo': 'val'})
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.contains, 'val')

    def test_apply_w_missing_name(self):
        from zerodbext.catalog.query import Name
        catalog = DummyCatalog()
        inst = self._makeOne('index', Name('foo'))
        self.assertRaises(NameError, inst._apply, catalog, {})

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "'val' in index")

    def test_negate(self):
        from zerodbext.catalog.query import DoesNotContain
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), DoesNotContain('index', 'val'))


class TestDoesNotContain(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import DoesNotContain
        return DoesNotContain

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.does_not_contain, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "'val' not in index")

    def test_negate(self):
        from zerodbext.catalog.query import Contains
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), Contains('index', 'val'))


class TestEq(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Eq
        return Eq

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.eq, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "index == 'val'")

    def test_negate(self):
        from zerodbext.catalog.query import NotEq
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), NotEq('index', 'val'))


class TestNotEq(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import NotEq
        return NotEq

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.not_eq, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "index != 'val'")

    def test_negate(self):
        from zerodbext.catalog.query import Eq
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), Eq('index', 'val'))


class TestGt(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Gt
        return Gt

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.gt, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "index > 'val'")

    def test_negate(self):
        from zerodbext.catalog.query import Le
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), Le('index', 'val'))


class TestLt(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Lt
        return Lt

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.lt, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "index < 'val'")

    def test_negate(self):
        from zerodbext.catalog.query import Ge
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), Ge('index', 'val'))


class TestGe(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Ge
        return Ge

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.ge, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "index >= 'val'")

    def test_negate(self):
        from zerodbext.catalog.query import Lt
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), Lt('index', 'val'))


class TestLe(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Le
        return Le

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.le, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', 'val')
        self.assertEqual(str(inst), "index <= 'val'")

    def test_negate(self):
        from zerodbext.catalog.query import Gt
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), Gt('index', 'val'))


class TestAll(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import All
        return All

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.all, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', [1, 2, 3])
        self.assertEqual(str(inst), "index in all([1, 2, 3])")

    def test_negate(self):
        from zerodbext.catalog.query import NotAll
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), NotAll('index', 'val'))


class TestNotAll(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import NotAll
        return NotAll

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.all, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', [1, 2, 3])
        self.assertEqual(str(inst), "index not in all([1, 2, 3])")

    def test_negate(self):
        from zerodbext.catalog.query import All
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), All('index', 'val'))


class TestAny(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Any
        return Any

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.any, 'val')

    def test_apply_with_list(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', ['one', 'two'])
        result = inst._apply(catalog, None)
        self.assertEqual(result, ['one', 'two'])
        self.assertEqual(catalog.index.any, ['one', 'two'])

    def test_apply_with_tuple(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', ('one', 'two'))
        result = inst._apply(catalog, None)
        self.assertEqual(result, ('one', 'two'))
        self.assertEqual(catalog.index.any, ('one', 'two'))
        
    def test_apply_with_names(self):
        from zerodbext.catalog.query import Name
        catalog = DummyCatalog()
        inst = self._makeOne('index', [Name('foo'), Name('bar')])
        result = inst._apply(catalog, names={'foo': 'one', 'bar': 'two'})
        self.assertEqual(result, ['one', 'two'])
        self.assertEqual(catalog.index.any, ['one', 'two'])

    def test_apply_with_names_none(self):
        from zerodbext.catalog.query import Name
        catalog = DummyCatalog()
        inst = self._makeOne('index', [Name('foo'), Name('bar')])
        self.assertRaises(NameError, inst._apply, catalog, None)

    def test_apply_with_names_in_tuple(self):
        from zerodbext.catalog.query import Name
        catalog = DummyCatalog()
        inst = self._makeOne('index', (Name('foo'), Name('bar')))
        result = inst._apply(catalog, names={'foo': 'one', 'bar': 'two'})
        self.assertEqual(result, ('one', 'two'))
        self.assertEqual(catalog.index.any, ('one', 'two'))

    def test_to_str(self):
        inst = self._makeOne('index', [1, 2, 3])
        self.assertEqual(str(inst), "index in any([1, 2, 3])")

    def test_negate(self):
        from zerodbext.catalog.query import NotAny
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), NotAny('index', 'val'))


class TestNotAny(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import NotAny
        return NotAny

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'val')
        result = inst._apply(catalog, None)
        self.assertEqual(result, 'val')
        self.assertEqual(catalog.index.not_any, 'val')

    def test_to_str(self):
        inst = self._makeOne('index', [1, 2, 3])
        self.assertEqual(str(inst), "index not in any([1, 2, 3])")

    def test_negate(self):
        from zerodbext.catalog.query import Any
        inst = self._makeOne('index', 'val')
        self.assertEqual(inst.negate(), Any('index', 'val'))


class TestInRange(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import InRange
        return InRange

    def _makeOne(self, index, begin, end,
                 begin_exclusive=False, end_exclusive=False):
        return self._getTargetClass()(
            index, begin, end, begin_exclusive, end_exclusive)

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'begin', 'end')
        result = inst._apply(catalog, None)
        self.assertEqual(result, ('begin', 'end', False, False))
        self.assertEqual(
            catalog.index.range, ('begin', 'end', False, False))

    def test_apply_w_names(self):
        from zerodbext.catalog.query import Name
        catalog = DummyCatalog()
        inst = self._makeOne('index', Name('foo'), Name('bar'))
        result = inst._apply(catalog, {'foo': 'begin', 'bar': 'end'})
        self.assertEqual(result, ('begin', 'end', False, False))
        self.assertEqual(
            catalog.index.range, ('begin', 'end', False, False))

    def test_apply_w_names_missing(self):
        from zerodbext.catalog.query import Name
        catalog = DummyCatalog()
        inst = self._makeOne('index', Name('foo'), Name('bar'))
        self.assertRaises(NameError, inst._apply, catalog, {})
        self.assertRaises(NameError, inst._apply, catalog, {'foo': 'begin'})

    def test_apply_exclusive(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'begin', 'end', True, True)
        result = inst._apply(catalog, None)
        self.assertEqual(result, ('begin', 'end', True, True))
        self.assertEqual(
            catalog.index.range, ('begin', 'end', True, True))

    def test_to_str(self):
        inst = self._makeOne('index', 0, 5)
        self.assertEqual(str(inst), "0 <= index <= 5")

    def test_to_str_exclusive(self):
        inst = self._makeOne('index', 0, 5, True, True)
        self.assertEqual(str(inst), "0 < index < 5")

    def test_from_gtlt(self):
        from zerodbext.catalog.query import Ge
        from zerodbext.catalog.query import Le
        gt = Ge('index', 0)
        lt = Le('index', 5)
        inst = self._getTargetClass().fromGTLT(gt, lt)
        self.assertEqual(str(inst), "0 <= index <= 5")

    def test_from_gtlt_exclusive(self):
        from zerodbext.catalog.query import Gt
        from zerodbext.catalog.query import Lt
        gt = Gt('index', 0)
        lt = Lt('index', 5)
        inst = self._getTargetClass().fromGTLT(gt, lt)
        self.assertEqual(str(inst), "0 < index < 5")

    def test_negate(self):
        from zerodbext.catalog.query import NotInRange
        inst = self._makeOne('index', 'begin', 'end')
        self.assertEqual(inst.negate(), NotInRange('index', 'begin', 'end'))

    def test_not_equal_to_another_type(self):
        inst = self._makeOne('index', 'begin', 'end')
        self.assertNotEqual(inst, object())


class TestNotInRange(ComparatorTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import NotInRange
        return NotInRange

    def _makeOne(self, index, begin, end,
                 begin_exclusive=False, end_exclusive=False):
        return self._getTargetClass()(
            index, begin, end, begin_exclusive, end_exclusive)

    def test_apply(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'begin', 'end')
        result = inst._apply(catalog, None)
        self.assertEqual(result, ('begin', 'end', False, False))
        self.assertEqual(
            catalog.index.not_range, ('begin', 'end', False, False))

    def test_apply_exclusive(self):
        catalog = DummyCatalog()
        inst = self._makeOne('index', 'begin', 'end', True, True)
        result = inst._apply(catalog, None)
        self.assertEqual(result, ('begin', 'end', True, True))
        self.assertEqual(
            catalog.index.not_range, ('begin', 'end', True, True))

    def test_to_str(self):
        inst = self._makeOne('index', 0, 5)
        self.assertEqual(str(inst), "not(0 <= index <= 5)")

    def test_to_str_exclusive(self):
        inst = self._makeOne('index', 0, 5, True, True)
        self.assertEqual(str(inst), "not(0 < index < 5)")

    def test_negate(self):
        from zerodbext.catalog.query import InRange
        inst = self._makeOne('index', 'begin', 'end')
        self.assertEqual(inst.negate(), InRange('index', 'begin', 'end'))

    def test_not_equal_to_another_type(self):
        inst = self._makeOne('index', 'begin', 'end')
        self.assertNotEqual(inst, object())


class BoolOpTestBase(unittest.TestCase):

    def _makeOne(self, left, right):
        return self._getTargetClass()(left, right)


class TestBoolOp(BoolOpTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import BoolOp as cls
        return cls

    def test_iter_children(self):

        class Dummy(object):
            pass

        left, right = Dummy(), Dummy()
        o = self._makeOne(left, right)
        self.assertEqual(list(o.iter_children()), [left, right])


class TestOr(BoolOpTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import Or as cls
        return cls

    def test_to_str(self):
        o = self._makeOne(None, None)
        self.assertEqual(str(o), 'Or')

    def test_apply(self):
        left = DummyQuery(set([1, 2]))
        right = DummyQuery(set([3, 4]))
        o = self._makeOne(left, right)
        o.family = DummyFamily()
        self.assertEqual(o._apply(None, None), set([1, 2, 3, 4]))
        self.assertTrue(left.applied)
        self.assertTrue(right.applied)
        self.assertEqual(o.family.union, (left.results, right.results))

    def test_apply_left_empty(self):
        left = DummyQuery(set())
        right = DummyQuery(set([3, 4]))
        o = self._makeOne(left, right)
        o.family = DummyFamily()
        self.assertEqual(o._apply(None, None), set([3, 4]))
        self.assertTrue(left.applied)
        self.assertTrue(right.applied)
        self.assertEqual(o.family.union, None)

    def test_apply_right_empty(self):
        left = DummyQuery(set([1, 2]))
        right = DummyQuery(set())
        o = self._makeOne(left, right)
        o.family = DummyFamily()
        self.assertEqual(o._apply(None, None), set([1, 2]))
        self.assertTrue(left.applied)
        self.assertTrue(right.applied)
        self.assertEqual(o.family.union, None)

    def test_negate(self):
        from zerodbext.catalog.query import And
        left = DummyQuery('foo')
        right = DummyQuery('bar')
        o = self._makeOne(left, right)
        neg = o.negate()
        self.assertTrue(isinstance(neg, And))
        left, right = neg.queries
        self.assertTrue(left.negated)
        self.assertTrue(right.negated)


class TestAnd(BoolOpTestBase):

    def _getTargetClass(self):
        from zerodbext.catalog.query import And as cls
        return cls

    def test_to_str(self):
        o = self._makeOne(None, None)
        self.assertEqual(str(o), 'And')

    def test_apply(self):
        left = DummyQuery(set([1, 2, 3]))
        right = DummyQuery(set([3, 4, 5]))
        o = self._makeOne(left, right)
        o.family = DummyFamily()
        self.assertEqual(o._apply(None, None), set([3]))
        self.assertTrue(left.applied)
        self.assertTrue(right.applied)
        self.assertEqual(o.family.intersection, (left.results, right.results))

    def test_apply_left_empty(self):
        left = DummyQuery(set([]))
        right = DummyQuery(set([3, 4, 5]))
        o = self._makeOne(left, right)
        o.family = DummyFamily()
        self.assertEqual(o._apply(None, None), set())
        self.assertTrue(left.applied)
        self.assertFalse(right.applied)
        self.assertEqual(o.family.intersection, None)

    def test_apply_right_empty(self):
        left = DummyQuery(set([1, 2, 3]))
        right = DummyQuery(set())
        o = self._makeOne(left, right)
        o.family = DummyFamily()
        self.assertEqual(o._apply(None, None), set())
        self.assertTrue(left.applied)
        self.assertTrue(right.applied)
        self.assertEqual(o.family.intersection, None)

    def test_negate(self):
        from zerodbext.catalog.query import Or
        left = DummyQuery('foo')
        right = DummyQuery('bar')
        o = self._makeOne(left, right)
        neg = o.negate()
        self.assertTrue(isinstance(neg, Or))
        left, right = neg.queries
        self.assertTrue(left.negated)
        self.assertTrue(right.negated)


class TestNot(BoolOpTestBase):

    def _makeOne(self, query):
        from zerodbext.catalog.query import Not as cls
        return cls(query)

    def test_to_str(self):
        o = self._makeOne(None)
        self.assertEqual(str(o), 'Not')

    def test_apply(self):
        query = DummyQuery('foo')
        o = self._makeOne(query)
        self.assertEqual(o._apply(None, None), 'foo')
        self.assertTrue(query.negated)
        self.assertTrue(query.applied)

    def test_negate(self):
        query = DummyQuery('foo')
        o = self._makeOne(query)
        self.assertEqual(o.negate(), query)

    def test_iter_children(self):
        query = DummyQuery('foo')
        o = self._makeOne(query)
        self.assertEqual(list(o.iter_children()), [query])


class TestName(unittest.TestCase):

    def _makeOne(self):
        from zerodbext.catalog.query import Name as cls
        return cls('foo')

    def test_to_str(self):
        o = self._makeOne()
        self.assertEqual(str(o), "Name('foo')")

    def test_eq(self):
        o1 = self._makeOne()
        o2 = self._makeOne()
        self.assertFalse(o1 is o2)
        self.assertTrue(o1 == o2)
        self.assertFalse(o1 == 'foo')


class Test_parse_query(unittest.TestCase):

    def tearDown(self):
        from zerodbext.catalog import query
        query.ast_support = True

    def _call_fut(self, expr):
        from zerodbext.catalog.query import parse_query as fut
        return fut(expr)

    def test_not_an_expression(self):
        self.assertRaises(ValueError, self._call_fut, 'a = 1')

    def test_multiple_expressions(self):
        self.assertRaises(ValueError, self._call_fut, 'a == 1\nb == 2\n')

    def test_unhandled_operator(self):
        self.assertRaises(ValueError, self._call_fut, 'a ^ b')

    def test_non_string_index_name(self):
        # == is not commutative in this context, sorry.
        self.assertRaises(ValueError, self._call_fut, '1 == a')

    def test_bad_operand_for_set_operation(self):
        self.assertRaises(ValueError, self._call_fut, '(a == 1) | 2')
        self.assertRaises(ValueError, self._call_fut, '1 | (b == 2)')

    def test_bad_operand_for_bool_operation(self):
        self.assertRaises(ValueError, self._call_fut, '1 or 2')

    def test_bad_comparator_chaining(self):
        self.assertRaises(ValueError, self._call_fut, '1 < 2 > 3')
        self.assertRaises(ValueError, self._call_fut, 'x == y == z')

    def test_bad_func_call(self):
        self.assertRaises(ValueError, self._call_fut, 'a in foo(bar)')

    def test_wrong_number_or_args_for_any(self):
        self.assertRaises(ValueError, self._call_fut, 'a in any(1, 2)')

    def test_no_ast_support(self):
        from zerodbext.catalog import query
        query.ast_support = False
        self.assertRaises(NotImplementedError, self._call_fut, None)

    def test_num(self):
        self.assertEqual(self._call_fut('1'), 1)
        self.assertEqual(self._call_fut('1.1'), 1.1)

    def test_str(self):
        self.assertEqual(self._call_fut('"foo"'), 'foo')

    def test_unicode(self):
        self.assertEqual(self._call_fut('u"foo"'), u'foo')

    def test_list(self):
        self.assertEqual(self._call_fut('[1, 2, 3]'), [1, 2, 3])

    def test_tuple(self):
        from zerodbext.catalog.query import Name
        self.assertEqual(self._call_fut('(a, b, c)'),
                         (Name('a'), Name('b'), Name('c')))

    def test_dotted_name(self):
        self.assertEqual(self._call_fut('a.foo').id, 'a.foo')

    def test_dotted_names(self):
        self.assertEqual(self._call_fut('a.foo.bar').id, 'a.foo.bar')

    def test_eq(self):
        from zerodbext.catalog.query import Eq
        eq = self._call_fut('a.foo == 1')
        self.assertTrue(isinstance(eq, Eq))
        self.assertEqual(eq.index_name, 'a.foo')
        self.assertEqual(eq._value, 1)

    def test_not_eq(self):
        from zerodbext.catalog.query import NotEq
        not_eq = self._call_fut("a != 'one'")
        self.assertTrue(isinstance(not_eq, NotEq))
        self.assertEqual(not_eq.index_name, 'a')
        self.assertEqual(not_eq._value, "one")

    def test_lt(self):
        from zerodbext.catalog.query import Lt
        from zerodbext.catalog.query import Name
        lt = self._call_fut("a < foo")
        self.assertTrue(isinstance(lt, Lt))
        self.assertEqual(lt.index_name, 'a')
        self.assertEqual(lt._value, Name('foo'))

    def test_le(self):
        from zerodbext.catalog.query import Le
        le = self._call_fut("a <= 4")
        self.assertTrue(isinstance(le, Le))
        self.assertEqual(le.index_name, 'a')
        self.assertEqual(le._value, 4)

    def test_gt(self):
        from zerodbext.catalog.query import Gt
        gt = self._call_fut('b > 2')
        self.assertTrue(isinstance(gt, Gt))
        self.assertEqual(gt.index_name, 'b')
        self.assertEqual(gt._value, 2)

    def test_ge(self):
        from zerodbext.catalog.query import Ge
        ge = self._call_fut("a >= 5")
        self.assertTrue(isinstance(ge, Ge))
        self.assertEqual(ge.index_name, 'a')
        self.assertEqual(ge._value, 5)

    def test_contains(self):
        from zerodbext.catalog.query import Contains
        contains = self._call_fut("6 in a")
        self.assertTrue(isinstance(contains, Contains))
        self.assertEqual(contains.index_name, 'a')
        self.assertEqual(contains._value, 6)

    def test_does_not_contain(self):
        from zerodbext.catalog.query import DoesNotContain
        contains = self._call_fut("6 not in a")
        self.assertTrue(isinstance(contains, DoesNotContain))
        self.assertEqual(contains.index_name, 'a')
        self.assertEqual(contains._value, 6)

    def test_range_exclusive_exclusive(self):
        from zerodbext.catalog.query import InRange
        comp = self._call_fut("0 < a < 5")
        self.assertTrue(isinstance(comp, InRange))
        self.assertEqual(comp.index_name, 'a')
        self.assertEqual(comp._start, 0)
        self.assertEqual(comp._end, 5)
        self.assertTrue(comp.start_exclusive)
        self.assertTrue(comp.end_exclusive)

    def test_range_exclusive_inclusive(self):
        from zerodbext.catalog.query import InRange
        comp = self._call_fut("0 < a <= 5")
        self.assertTrue(isinstance(comp, InRange))
        self.assertEqual(comp.index_name, 'a')
        self.assertEqual(comp._start, 0)
        self.assertEqual(comp._end, 5)
        self.assertTrue(comp.start_exclusive)
        self.assertFalse(comp.end_exclusive)

    def test_range_inclusive_exclusive(self):
        from zerodbext.catalog.query import InRange
        comp = self._call_fut("0 <= a < 5")
        self.assertTrue(isinstance(comp, InRange))
        self.assertEqual(comp.index_name, 'a')
        self.assertEqual(comp._start, 0)
        self.assertEqual(comp._end, 5)
        self.assertFalse(comp.start_exclusive)
        self.assertTrue(comp.end_exclusive)

    def test_range_inclusive_inclusive(self):
        from zerodbext.catalog.query import InRange
        comp = self._call_fut("0 <= a <= 5")
        self.assertTrue(isinstance(comp, InRange))
        self.assertEqual(comp.index_name, 'a')
        self.assertEqual(comp._start, 0)
        self.assertEqual(comp._end, 5)
        self.assertFalse(comp.start_exclusive)
        self.assertFalse(comp.end_exclusive)

    def test_not_in_range(self):
        from zerodbext.catalog.query import NotInRange
        comp = self._call_fut("not(0 < a < 5)")
        self.assertTrue(isinstance(comp, NotInRange))
        self.assertEqual(comp.index_name, 'a')
        self.assertEqual(comp._start, 0)
        self.assertEqual(comp._end, 5)
        self.assertTrue(comp.start_exclusive)
        self.assertTrue(comp.end_exclusive)

    def test_or(self):
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import Or
        op = self._call_fut("(a == 1) | (b == 2)")
        self.assertTrue(isinstance(op, Or))
        query = op.queries[0]
        self.assertTrue(isinstance(query, Eq))
        self.assertEqual(query.index_name, 'a')
        self.assertEqual(query._value, 1)
        query = op.queries[1]
        self.assertTrue(isinstance(query, Eq))
        self.assertEqual(query.index_name, 'b')
        self.assertEqual(query._value, 2)

    def test_or_with_bool_syntax(self):
        from zerodbext.catalog.query import NotEq
        from zerodbext.catalog.query import Or
        op = self._call_fut("a != 1 or b != 2")
        self.assertTrue(isinstance(op, Or))
        query = op.queries[0]
        self.assertTrue(isinstance(query, NotEq))
        self.assertEqual(query.index_name, 'a')
        self.assertEqual(query._value, 1)
        query = op.queries[1]
        self.assertTrue(isinstance(query, NotEq))
        self.assertEqual(query.index_name, 'b')
        self.assertEqual(query._value, 2)

    def test_any(self):
        from zerodbext.catalog.query import Any
        op = self._call_fut("a == 1 or a == 2 or a == 3")
        self.assertTrue(isinstance(op, Any), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_better_any(self):
        from zerodbext.catalog.query import Any
        op = self._call_fut("a in any([1, 2, 3])")
        self.assertTrue(isinstance(op, Any), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_any_with_name(self):
        from zerodbext.catalog.query import Any
        from zerodbext.catalog.query import Name
        op = self._call_fut("a in any(foo)")
        self.assertTrue(isinstance(op, Any), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, Name('foo'))

    def test_any_with_names(self):
        from zerodbext.catalog.query import Any
        from zerodbext.catalog.query import Name
        op = self._call_fut("a in any([foo, bar])")
        self.assertTrue(isinstance(op, Any), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [Name('foo'), Name('bar')])

    def test_not_any(self):
        from zerodbext.catalog.query import NotAny
        op = self._call_fut("not(a == 1 or a == 2 or a == 3)")
        self.assertTrue(isinstance(op, NotAny), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_better_not_any(self):
        from zerodbext.catalog.query import NotAny
        op = self._call_fut("a not in any([1, 2, 3])")
        self.assertTrue(isinstance(op, NotAny), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_and(self):
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import And
        op = self._call_fut("(a == 1) & (b == 2)")
        self.assertTrue(isinstance(op, And))
        query = op.queries[0]
        self.assertTrue(isinstance(query, Eq))
        self.assertEqual(query.index_name, 'a')
        self.assertEqual(query._value, 1)
        query = op.queries[1]
        self.assertTrue(isinstance(query, Eq))
        self.assertEqual(query.index_name, 'b')
        self.assertEqual(query._value, 2)

    def test_and_with_bool_syntax(self):
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import And
        op = self._call_fut("a == 1 and b == 2")
        self.assertTrue(isinstance(op, And))
        query = op.queries[0]
        self.assertTrue(isinstance(query, Eq))
        self.assertEqual(query.index_name, 'a')
        self.assertEqual(query._value, 1)
        query = op.queries[1]
        self.assertTrue(isinstance(query, Eq))
        self.assertEqual(query.index_name, 'b')
        self.assertEqual(query._value, 2)

    def test_all(self):
        from zerodbext.catalog.query import All
        op = self._call_fut("a == 1 and a == 2 and a == 3")
        self.assertTrue(isinstance(op, All), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_better_all(self):
        from zerodbext.catalog.query import All
        op = self._call_fut("a in all([1, 2, 3])")
        self.assertTrue(isinstance(op, All), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_not_all(self):
        from zerodbext.catalog.query import NotAll
        op = self._call_fut("not(a == 1 and a == 2 and a == 3)")
        self.assertTrue(isinstance(op, NotAll), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_better_not_all(self):
        from zerodbext.catalog.query import NotAll
        op = self._call_fut("a not in all([1, 2, 3])")
        self.assertTrue(isinstance(op, NotAll), op)
        self.assertEqual(op.index_name, 'a')
        self.assertEqual(op._value, [1, 2, 3])

    def test_all_with_or(self):
        # Regression test for earlier bug where:
        #   a == 1 or a == 2 and a == 3
        # was transformed into:
        #   a any [1, 2, 3]
        from zerodbext.catalog.query import All
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import Or
        op = self._call_fut("a == 1 or a == 2 and a == 3")
        self.assertTrue(isinstance(op, Or))
        self.assertTrue(isinstance(op.queries[0], Eq))
        self.assertTrue(isinstance(op.queries[1], All))
        self.assertEqual(op.queries[1].index_name, 'a')
        self.assertEqual(op.queries[1]._value, [2, 3])

    def test_convert_gtlt_to_range(self):
        from zerodbext.catalog.query import InRange
        op = self._call_fut("a < 1 and a > 0")
        self.assertTrue(isinstance(op, InRange))
        self.assertEqual(op._start, 0)
        self.assertEqual(op._end, 1)
        self.assertEqual(op.start_exclusive, True)
        self.assertEqual(op.end_exclusive, True)

    def test_convert_ltgt_to_range(self):
        from zerodbext.catalog.query import InRange
        op = self._call_fut("a > 0 and a < 1")
        self.assertTrue(isinstance(op, InRange))
        self.assertEqual(op._start, 0)
        self.assertEqual(op._end, 1)
        self.assertEqual(op.start_exclusive, True)
        self.assertEqual(op.end_exclusive, True)

    def test_convert_gtlt_to_not_in_range(self):
        from zerodbext.catalog.query import NotInRange
        op = self._call_fut("a < 0 or a > 1")
        self.assertTrue(isinstance(op, NotInRange))
        self.assertEqual(op._start, 0)
        self.assertEqual(op._end, 1)
        self.assertEqual(op.start_exclusive, False)
        self.assertEqual(op.end_exclusive, False)

    def test_convert_ltgt_to_not_in_range(self):
        from zerodbext.catalog.query import NotInRange
        op = self._call_fut("a > 1 or a < 0")
        self.assertTrue(isinstance(op, NotInRange))
        self.assertEqual(op._start, 0)
        self.assertEqual(op._end, 1)
        self.assertEqual(op.start_exclusive, False)
        self.assertEqual(op.end_exclusive, False)

    def test_convert_gtlt_child_left_nephew_left(self):
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import And
        from zerodbext.catalog.query import InRange
        op = self._call_fut("a > 0 and (a < 5 and b == 7)")
        self.assertTrue(isinstance(op, And))
        self.assertTrue(isinstance(op.queries[0], InRange))
        self.assertTrue(isinstance(op.queries[1], Eq))

    def test_strange_gtlt_child_left_nephew_right(self):
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import And
        from zerodbext.catalog.query import InRange
        op = self._call_fut("a > 0 and (b == 7 and a < 5)")
        self.assertTrue(isinstance(op, And))
        self.assertTrue(isinstance(op.queries[0], InRange))
        self.assertTrue(isinstance(op.queries[1], Eq))

    def test_convert_gtlt_child_right_nephew_left(self):
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import Gt
        from zerodbext.catalog.query import And
        from zerodbext.catalog.query import InRange
        op = self._call_fut("a >= -1 and b == 2 and c > 3 and a <= 1")
        self.assertTrue(isinstance(op, And))
        self.assertTrue(isinstance(op.queries[0], InRange))
        self.assertTrue(isinstance(op.queries[1], Eq))
        self.assertTrue(isinstance(op.queries[2], Gt))

    def test_convert_gtlt_both_descendants(self):
        from zerodbext.catalog.query import Eq
        from zerodbext.catalog.query import Gt
        from zerodbext.catalog.query import And
        from zerodbext.catalog.query import InRange
        op = self._call_fut("b == 2 and a > -1 and (a <= 1 and c > 3)")
        self.assertTrue(isinstance(op, And))
        self.assertTrue(isinstance(op.queries[0], Eq))
        self.assertTrue(isinstance(op.queries[1], InRange))
        self.assertTrue(isinstance(op.queries[2], Gt))

    def test_convert_gtlt_both_descendants_multiple_times(self):
        from zerodbext.catalog.query import And
        from zerodbext.catalog.query import InRange
        op = self._call_fut(
            "(a > 0 and b > 0 and c > 0) and (a < 5 and b < 5 and c < 5)")
        self.assertTrue(isinstance(op, And))
        self.assertTrue(isinstance(op.queries[0], InRange))
        self.assertTrue(isinstance(op.queries[1], InRange))
        self.assertTrue(isinstance(op.queries[2], InRange))

    def test_dont_convert_gtlt_to_range_with_or_spread_out(self):
        from zerodbext.catalog.query import Gt
        from zerodbext.catalog.query import Lt
        from zerodbext.catalog.query import And
        from zerodbext.catalog.query import Or
        op = self._call_fut("a > 0 and b > 0 or a < 5 and b < 5")
        self.assertTrue(isinstance(op, Or))
        self.assertTrue(isinstance(op.queries[0], And))
        self.assertTrue(isinstance(op.queries[0].queries[0], Gt))
        self.assertTrue(isinstance(op.queries[0].queries[1], Gt))
        self.assertTrue(isinstance(op.queries[1], And))
        self.assertTrue(isinstance(op.queries[1].queries[0], Lt))
        self.assertTrue(isinstance(op.queries[1].queries[1], Lt))


if not ast_support:  # pragma NO COVERAGE
    del Test_parse_query


class Dummy(object):
    __parent__ = None


class DummyCatalog(object):

    def __init__(self, index=None):
        if index is None:
            index = DummyIndex()
        self.index = index

    def __getitem__(self, name):
        return self.index


class DummyIndex(object):

    def applyContains(self, value):
        self.contains = value
        return value

    def applyDoesNotContain(self, value):
        self.does_not_contain = value
        return value

    def applyEq(self, value):
        self.eq = value
        return value

    def applyNotEq(self, value):
        self.not_eq = value
        return value

    def applyGt(self, value):
        self.gt = value
        return value

    def applyLt(self, value):
        self.lt = value
        return value

    def applyGe(self, value):
        self.ge = value
        return value

    def applyLe(self, value):
        self.le = value
        return value

    def applyAny(self, value):
        self.any = value
        return value

    def applyNotAny(self, value):
        self.not_any = value
        return value

    def applyAll(self, value):
        self.all = value
        return value

    def applyInRange(self, start, end, start_exclusive, end_exclusive):
        self.range = (start, end, start_exclusive, end_exclusive)
        return self.range

    def applyNotInRange(self, start, end, start_exclusive, end_exclusive):
        self.not_range = (start, end, start_exclusive, end_exclusive)
        return self.not_range


class DummyFamily(object):
    union = None
    intersection = None
    diff = None

    @property
    def IF(self):
        return self

    def weightedUnion(self, left, right):
        self.union = (left, right)
        return None, left | right

    def weightedIntersection(self, left, right):
        self.intersection = (left, right)
        return None, left & right

    def Set(self):
        return set()


class DummyQuery(object):
    applied = False
    negated = False

    def __init__(self, results):
        self.results = results

    def _apply(self, catalog, names):
        self.applied = True
        return self.results

    def negate(self):
        self.negated = True
        return self



import random

from persistent import Persistent
from BTrees.IOBTree import IOBTree
from BTrees.OIBTree import OIBTree
from BTrees.OOBTree import OOBTree

import BTrees

_marker = ()

class DocumentMap(Persistent):
    """ A two-way map between addresses (e.g. location paths) and document ids.

    The map is a persistent object meant to live in a ZODB storage.

    Additionally, the map is capable of mapping 'metadata' to docids.
    """
    _v_nextid = None
    family = BTrees.family32
    _randrange = random.randrange
    docid_to_metadata = None # latch for b/c

    def __init__(self):
        self.docid_to_address = IOBTree()
        self.address_to_docid = OIBTree()
        self.docid_to_metadata = IOBTree()

    def docid_for_address(self, address):
        """ Retrieve a document id for a given address.

        ``address`` is a string or other hashable object which represents
        a token known by the application.

        Return the integer document id corresponding to ``address``.

        If ``address`` doesn't exist in the document map, return None.
        """
        return self.address_to_docid.get(address)

    def address_for_docid(self, docid):
        """ Retrieve an address for a given document id.

        ``docid`` is an integer document id.

        Return the address corresponding to ``docid``.

        If ``docid`` doesn't exist in the document map, return None.
        """
        return self.docid_to_address.get(docid)

    def add(self, address, docid=_marker):
        """ Add a new document to the document map.

        ``address`` is a string or other hashable object which represents
        a token known by the application.

        ``docid``, if passed, must be an int.  In this case, remove
        any previous address stored for it before mapping it to the
        new address.  Passing an explicit ``docid`` also removes any
        metadata associated with that docid.
        
        If ``docid`` is not passed, generate a new docid.

        Return the integer document id mapped to ``address``.
        """
        if docid is _marker:
            docid = self.new_docid()

        self.remove_docid(docid)
        self.remove_address(address)

        self.docid_to_address[docid] = address
        self.address_to_docid[address] = docid
        return docid

    def remove_docid(self, docid):
        """ Remove a document from the document map for the given document ID.

        ``docid`` is an integer document id.

        Remove any corresponding metadata for ``docid`` as well.

        Return a True if ``docid`` existed in the map, else return False.
        """
        # It should be an invariant that if one entry exists in
        # docid_to_address for a docid/address pair, exactly one
        # corresponding entry exists in address_to_docid for the same
        # docid/address pair.  However, versions of this code before
        # r.catalog 0.7.3 had a bug which, if this method was called
        # multiple times, each time with the same address but a
        # different docid, the ``docid_to_address`` mapping could
        # contain multiple entries for the same address each with a
        # different docid, causing this invariant to be violated.  The
        # symptom: in systems that used r.catalog 0.7.2 and lower,
        # there might be more entries in docid_to_address than there
        # are in address_to_docid.  The conditional fuzziness in the
        # code directly below is a runtime kindness to systems in that
        # state.  Technically, the administrator of a system in such a
        # state should normalize the two data structures by running a
        # script after upgrading to 0.7.3.  If we made the admin do
        # this, some of the code fuzziness below could go away,
        # replaced with something simpler.  But there's no sense in
        # breaking systems at runtime through being a hardass about
        # consistency if an unsuspecting upgrader has not yet run the
        # data fixer script. The "fix the data" mantra rings a
        # little hollow when you weren't the one who broke the data in
        # the first place ;-)

        self._check_metadata()

        address = self.docid_to_address.get(docid, _marker)
        if address is _marker:
            return False
        
        old_docid = self.address_to_docid.get(address, _marker)
        if (old_docid is not _marker) and (old_docid != docid):
            self.remove_docid(old_docid)

        if docid in self.docid_to_address:
            del self.docid_to_address[docid]
        if address in self.address_to_docid:
            del self.address_to_docid[address]
        if docid in self.docid_to_metadata:
            del self.docid_to_metadata[docid]

        return True


    def remove_address(self, address):
        """ Remove a document from the document map using an address.

        ``address`` is a string or other hashable object which represents
        a token known by the application.

        Remove any corresponding metadata for ``address`` as well.

        Return a True if ``address`` existed in the map, else return False.
        """
        # See the comment in remove_docid for complexity rationalization
        
        self._check_metadata()

        docid = self.address_to_docid.get(address, _marker)
        if docid is _marker:
            return False
        
        old_address = self.docid_to_address.get(docid, _marker)
        if (old_address is not _marker) and (old_address != address):
            self.remove_address(old_address)

        if docid in self.docid_to_address:
            del self.docid_to_address[docid]
        if address in self.address_to_docid:
            del self.address_to_docid[address]
        if docid in self.docid_to_metadata:
            del self.docid_to_metadata[docid]

        return True

    def _check_metadata(self):
        # backwards compatibility
        if self.docid_to_metadata is None:
            self.docid_to_metadata = IOBTree()

    def add_metadata(self, docid, data):
        """ Add metadata related to a given document id.

        ``data`` must be a mapping, such as a dictionary.
        
        For each key/value pair in ``data`` insert a metadata key/value pair
        into the metadata stored for ``docid``.

        Overwrite any existing values for the keys in ``data``, leaving values
        unchanged for other existing keys.

        Raise a KeyError If ``docid`` doesn't relate to an address in the
        document map.
        """
        if not docid in self.docid_to_address:
            raise KeyError(docid)
        if len(data.keys()) == 0:
            return
        self._check_metadata()
        meta = self.docid_to_metadata.setdefault(docid, OOBTree())
        for k in data:
            meta[k] = data[k]

    def remove_metadata(self, docid, *keys):
        """ Remove metadata related to a given document id.

        If ``docid`` doesn't exist in the metadata map, raise a KeyError.

        For each key in ``keys``, remove the metadata value for the
        docid related to that key.
        
        Do not raise any error if no value exists for a given key.

        If no keys are specified, remove all metadata related to the docid.
        """
        self._check_metadata()
        if keys:
            meta = self.docid_to_metadata.get(docid, _marker)
            if meta is _marker:
                raise KeyError(docid)
            for k in keys:
                if k in meta:
                    del meta[k]
            if not meta:
                del self.docid_to_metadata[docid]
        else:
            if not (docid in self.docid_to_metadata):
                raise KeyError(docid)
            del self.docid_to_metadata[docid]

    def get_metadata(self, docid):
        """ Return the metadata for ``docid``.

        Return a mapping of the keys and values set using ``add_metadata``.

        Raise a KeyError If metadata does not exist for ``docid``.
        """
        if self.docid_to_metadata is None:
            raise KeyError(docid)
        meta = self.docid_to_metadata[docid]
        return meta

    def new_docid(self):
        """ Return a new document id.

        The returned value is guaranteed not to be used already in this
        document map.
        """
        while True:
            if self._v_nextid is None:
                self._v_nextid = self._randrange(self.family.minint,
                                                 self.family.maxint)
            uid = self._v_nextid
            self._v_nextid += 1
            if uid not in self.docid_to_address:
                return uid
            self._v_nextid = None



try:
    from hashlib import md5
except: # pragma no cover
    from md5 import new as md5

import six
from persistent import Persistent
from zope.interface import implementer

from zerodbext.catalog.indexes.keyword import CatalogKeywordIndex
from zerodbext.catalog.interfaces import ICatalogIndex

_marker = ()


@implementer(ICatalogIndex)
class CatalogFacetIndex(CatalogKeywordIndex):
    """Facet index.

    Query types supported:

    - Eq

    - NotEq

    - In

    - NotIn

    - Any

    - NotAny

    - All

    - NotAll
    """

    def __init__(self, discriminator, facets, family=None):
        if not callable(discriminator):
            if not isinstance(discriminator, six.string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        if family is not None:
            self.family = family
        self.facets = self.family.OO.Set(facets)
        self._not_indexed = self.family.IF.Set()
        self.clear()

    def index_doc(self, docid, object):
        """ Pass in an integer document id and an object supporting a
        sequence of facet specifiers ala ['style:gucci:handbag'] via
        the discriminator"""

        if callable(self.discriminator):
            value = self.discriminator(object, _marker)
        else:
            value = getattr(object, self.discriminator, _marker)

        if value is _marker:
            # unindex the previous value
            self.unindex_doc(docid)
            self._not_indexed.add(docid)
            return None

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object %s' %
                             value)

        if docid in self._not_indexed:
            self._not_indexed.remove(docid)

        old = self._rev_index.get(docid)
        if old is not None:
            self.unindex_doc(docid)

        changed = False

        for facet in value:
            L = []
            categories = facet.split(':')
            for category in categories:
                L.append(category)
                facet_candidate = ':'.join(L)
                for fac in self.facets:
                    if fac == facet_candidate:
                        changed = True
                        fwset = self._fwd_index.get(fac)
                        if fwset is None:
                            fwset = self.family.IF.Set()
                            self._fwd_index[fac] = fwset
                        fwset.insert(docid)
                        revset = self._rev_index.get(docid)
                        if revset is None:
                            revset = self.family.OO.Set()
                            self._rev_index[docid] = revset
                        revset.insert(fac)

        if changed:
            self._num_docs.change(1)

        return value

    def counts(self, docids, omit_facets=()):
        """ Given a set of docids (usually returned from query),
        provide count information for further facet narrowing.
        Optionally omit count information for facets and their
        ancestors that are in 'omit_facets' (a sequence of facets)"""

        effective_omits = self.family.OO.Set()

        for omit_facet in omit_facets:
            L = []
            categories = omit_facet.split(':')
            for category in categories:
                L.append(category)
                effective_omits.insert(':'.join(L))

        include_facets = self.family.OO.difference(self.facets,
                                                   effective_omits)

        counts = {}
        isect_cache = {}

        for docid in docids:
            available_facets = self._rev_index.get(docid)
            ck = cachekey(available_facets)
            appropriate_facets = isect_cache.get(ck)
            if appropriate_facets is None:
                appropriate_facets = self.family.OO.intersection(
                    include_facets, available_facets)
                isect_cache[ck] = appropriate_facets
            for facet in appropriate_facets:
                count = counts.get(facet, 0)
                count += 1
                counts[facet] = count

        return counts


def cachekey(set):
    h = md5()
    for item in sorted(list(set)):
        h.update(item.encode())
    return h.hexdigest()




import unittest

_marker = object()


class TestCatalogKeywordIndex(unittest.TestCase):

    def _getTargetClass(self):
        from zerodbext.catalog.indexes.keyword import CatalogKeywordIndex
        return CatalogKeywordIndex

    def _makeOne(self, discriminator=_marker):

        def _discriminator(obj, default):
            if obj is _marker:
                return default
            return obj

        if discriminator is _marker:
            discriminator = _discriminator
        return self._getTargetClass()(discriminator)

    def test_class_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyClass
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyClass(ICatalogIndex, self._getTargetClass())

    def test_instance_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyObject
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyObject(ICatalogIndex, self._makeOne())

    def test_ctor_callback_discriminator(self):

        def _discriminator(obj, default):
            """ """

        index = self._makeOne(_discriminator)
        self.assertTrue(index.discriminator is _discriminator)

    def test_ctor_string_discriminator(self):
        index = self._makeOne('abc')
        self.assertEqual(index.discriminator, 'abc')

    def test_ctor_bad_discriminator(self):
        self.assertRaises(ValueError, self._makeOne, object())

    def test_reindex_doc_doesnt_unindex(self):
        index = self._makeOne()
        index.index_doc(5, [1])
        index.unindex_doc = lambda *args, **kw: 1 / 0
        index.reindex_doc(5, [1])

    def test_reindex_doc_same_values(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        self.assertEqual(index.documentCount(), 1)
        index.reindex_doc(1, [1, 2, 3])
        self.assertEqual(index.documentCount(), 1)
        self.assertTrue(1 in index._rev_index)
        self.assertTrue(1 in index._fwd_index[1])
        self.assertTrue(1 in index._fwd_index[2])
        self.assertTrue(1 in index._fwd_index[3])
        self.assertFalse(4 in index._fwd_index)

    def test_reindex_doc_different_values(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        self.assertEqual(index.documentCount(), 1)
        index.reindex_doc(1, [2, 3, 4])
        self.assertEqual(index.documentCount(), 1)
        self.assertTrue(1 in index._rev_index)
        self.assertFalse(1 in index._fwd_index.get(1, []))
        self.assertTrue(1 in index._fwd_index[2])
        self.assertTrue(1 in index._fwd_index[3])
        self.assertTrue(1 in index._fwd_index[4])

    def test_apply_doesnt_mutate_query(self):
        # Some previous version of zope.index munged the query dict
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        query = {'operator': 'or', 'query': [5]}
        result = index.apply(FrozenDict(query))
        self.assertEqual(list(result), [2, 3])
        self.assertEqual(query, {'operator': 'or', 'query': [5]})

    def test_applyAny(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        result = index.applyAny([5, 9])
        self.assertEqual(list(result), [2, 3, 4, 5])

    def test_applyNotAny(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        result = index.applyNotAny([5, 9])
        self.assertEqual(list(result), [1])

    def test_applyAll(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        result = index.applyAll([5, 9])
        self.assertEqual(list(result), [])

    def test_applyNotAll(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        result = index.applyNotAll([5, 9])
        self.assertEqual(list(result), [1, 2, 3, 4, 5])

    def test_applyEq(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        result = index.applyEq(5)
        self.assertEqual(list(result), [2, 3])

    def test_applyNotEq(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        result = index.applyNotEq(5)
        self.assertEqual(list(result), [1, 4, 5])

    def test_applyNotEq_with_unindexed_docs(self):

        def discriminator(obj, default):
            if isinstance(obj, list):
                return obj
            return default

        index = self._makeOne(discriminator)
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        index.index_doc(6, (5, 6))
        result = index.applyNotEq(5)
        self.assertEqual(list(result), [1, 4, 5, 6])

    def test_applyNotEq_nothing_indexed(self):
        def discriminator(obj, default):
            return default
        index = self._makeOne(discriminator)
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        index.index_doc(6, (5, 6))
        result = index.applyNotEq(5)
        self.assertEqual(list(result), [1, 2, 3, 4, 5, 6])

    def test_docids(self):
        index = self._makeOne()
        index.index_doc(1, [1, 2, 3])
        index.index_doc(2, [3, 4, 5])
        index.index_doc(3, [5, 6, 7])
        index.index_doc(4, [7, 8, 9])
        index.index_doc(5, [9, 10])
        index.index_doc(6, (5, 6))
        self.assertEqual(set(index.docids()),
                         set((1, 2, 3, 4, 5, 6)))

    def test_unindex_doc_removes_from_docids(self):
        index = self._makeOne()
        index.index_doc(20, [1, 2, 3])
        self.assertTrue(20 in index.docids())
        index.unindex_doc(20)
        self.assertFalse(20 in index.docids())

    def test_index_doc_then_missing_value(self):
        index = self._makeOne()
        index.index_doc(20, [1, 2, 3])
        self.assertEqual(set([20]), set(index.applyEq(2)))
        self.assertTrue(20 in index.docids())
        index.index_doc(20, _marker)
        self.assertEqual(set(), set(index.applyEq(2)))
        self.assertTrue(20 in index.docids())

    def test_index_doc_missing_value_then_with_value(self):
        index = self._makeOne()
        index.index_doc(3, _marker)
        self.assertEqual(set(), set(index.applyEq(4)))
        self.assertTrue(3 in index.docids())
        index.index_doc(3, [3, 4, 5])
        self.assertEqual(set([3]), set(index.applyEq(4)))
        self.assertTrue(3 in index.docids())

    def test_index_doc_missing_value_then_unindex(self):
        index = self._makeOne()
        index.index_doc(3, _marker)
        self.assertEqual(set(), set(index.applyEq(4)))
        self.assertTrue(3 in index.docids())
        index.unindex_doc(3)
        self.assertEqual(set(), set(index.applyEq(4)))
        self.assertFalse(3 in index.docids())

    def test_docids_with_indexed_and_not_indexed(self):
        index = self._makeOne()
        index.index_doc(1, [1])
        index.index_doc(2, _marker)
        self.assertEqual(set([1, 2]), set(index.docids()))


class FrozenDict(dict):
    def _forbidden(self, *args, **kw):
        assert 0 # pragma: no cover
    __setitem__ = __delitem__ = clear = update = _forbidden



from zope.interface import Interface
from zope.interface import Attribute

from zope.index.interfaces import IInjection
from zope.index.interfaces import IIndexSearch

class ICatalog(IIndexSearch, IInjection):
    def search(**query):
        """Search on the query provided.  Each query key is an index
        name, each query value is the value that the index expects as
        a query term."""

class ICatalogIndex(IIndexSearch, IInjection):
    """ An index that adapts objects to an attribute or callable value
    on an object """
    def apply_intersect(query, docids):
        """ Run the query implied by query, and return query results
        intersected with the ``docids`` set that is supplied.  If
        ``docids`` is None, return the bare query results. """

    def reindex_doc(docid, obj):
        """ Reindex the document numbered ``docid`` using in the
        information on object ``obj``"""

class ICatalogAdapter(Interface):
    def __call__(default):
        """ Return the value or the default if the object no longer
        has any value for the adaptation"""



import unittest


class TestMigration(unittest.TestCase):

    def test_migrate_to_0_8_0(self):
        from zerodbext.catalog.migration import migrate_to_0_8_0
        catalog = DummyCatalog()
        migrate_to_0_8_0(catalog)
        self.assertEqual(catalog['one'].migrated, set([1, 2, 3]))
        self.assertEqual(catalog['two'].migrated, set([1, 2, 3]))
        self.assertEqual(catalog['three'].migrated, False)


    def test_migrate_to_0_8_0_from_document_map(self):
        from zerodbext.catalog.migration import migrate_to_0_8_0_from_document_map
        catalog = DummyCatalog()
        migrate_to_0_8_0_from_document_map(catalog, catalog.document_map)
        self.assertEqual(catalog['one'].migrated, set([2, 3, 4]))
        self.assertEqual(catalog['two'].migrated, set([2, 3, 4]))
        self.assertEqual(catalog['three'].migrated, False)


class DummyCatalog(dict):

    def __init__(self):
        super(DummyCatalog, self).__init__()
        self['one'] = MigratableDummyIndex(1, 2)
        self['two'] = MigratableDummyIndex(2, 3)
        self['three'] = DummyIndex(3, 4)
        self.document_map = DummyDocumentMap()


class DummyIndex(object):
    migrated = False

    def __init__(self, *docids):
        self._docids = set(docids)


class MigratableDummyIndex(DummyIndex):

    def _migrate_to_0_8_0(self, docids):
        self.migrated = set(docids)

    def _indexed(self):
        return self._docids


class DummyDocumentMap(object):

    def __init__(self):
        self.docid_to_address = {2: '/one', 3: '/one/two', 4: '/three'}



import six
from zope.interface import implementer

from zope.index.interfaces import IIndexSort
from zope.index.text import TextIndex

from zerodbext.catalog.interfaces import ICatalogIndex
from zerodbext.catalog.indexes.common import CatalogIndex


@implementer(ICatalogIndex, IIndexSort)
class CatalogTextIndex(CatalogIndex, TextIndex):
    """ Full-text index.

    Query types supported:

    - Contains

    - DoesNotContain

    - Eq

    - NotEq
    """

    def __init__(self, discriminator, lexicon=None, index=None):
        if not callable(discriminator):
            if not isinstance(discriminator, six.string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        self._not_indexed = self.family.IF.Set()
        TextIndex.__init__(self, lexicon, index)
        self.clear()

    def reindex_doc(self, docid, object):
        # index_doc knows enough about reindexing to do the right thing
        return self.index_doc(docid, object)

    def _indexed(self):
        return self.index._docwords.keys()

    def sort(self, result, reverse=False, limit=None, sort_type=None):
        """Sort by text relevance.

        This only works if the query includes at least one text query,
        leading to a weighted result.  This method raises TypeError
        if the result is not weighted.

        A weighted result is a dictionary-ish object that has docids
        as keys and floating point weights as values.  This method
        sorts the dictionary by weight and returns the sorted
        docids as a list.
        """
        if not result:
            return result

        if not hasattr(result, 'items'):
            raise TypeError(
                "Unable to sort by relevance because the search "
                "result does not contain weights. To produce a weighted "
                "result, include a text search in the query.")

        items = [(weight, docid) for (docid, weight) in result.items()]
        # when reverse is false, output largest weight first.
        # when reverse is true, output smallest weight first.
        items.sort(reverse=not reverse)
        result = [docid for (weight, docid) in items]
        if limit:
            result = result[:limit]
        return result

    def applyContains(self, value):
        return self.apply(value)

    applyEq = applyContains



__version__ = '0.8.4'

import os

from setuptools import setup, find_packages

try:
    here = os.path.abspath(os.path.dirname(__file__))
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
except IOError:
    README = CHANGES = ''

INSTALL_REQUIRES = [
    'setuptools',
    'zope.component',
    'ZODB',
    'zope.index >= 3.5.0',
    'six'
]

TESTING_EXTRAS = ['nose', 'coverage']
DOCS_EXTRAS = ['Sphinx']

setup(name='zerodbext.catalog',
      version=__version__,
      description='Searching and indexing based on zope.index. Fork of repoze.catalog',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        ],
      keywords='indexing catalog search',
      author="ZeroDB",
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      packages=find_packages(),
      include_package_data=True,
      namespace_packages=['zerodbext'],
      zip_safe=False,
      tests_require = INSTALL_REQUIRES,
      install_requires = INSTALL_REQUIRES,
      extras_require = {
            'benchmark': ['PyChart'],
            'testing': TESTING_EXTRAS,
            'docs': DOCS_EXTRAS,
        },
      test_suite="zerodbext.catalog",
)



"""
Version 0.8.0 introduced a NOT operator which required that indexes keep track
of some data they hadn't previously needed to keep track of. Previously, if a
discriminator returned `default` when indexing a document, the index did not
keep track of the fact that it had seen that docid. In order to properly
implement query negations via the NOT operator, indexes now keep track of all
docids which have been indexed regardless of whether those docids had a value
for the index.

The functions in this module will bring indices in an existing catalog up to
date with this information. If you are using a `DocumentMap`, the easiest way
to migrate your indices will be to call `migrate_to_0_8_0_from_document_map`.
If you are not using a `DocumentMap` but otherwise have a convenient means of
generating the complete set of docids in your catalog, you can use
`migrate_to_0_8_0_from_docids`. Failing both of the previous options, you can
call `migrate_to_0_8_0`, which will try to build up a set of all docids in the
catalog by looking at all of the indexes first.
"""

import BTrees
IF = BTrees.family32.IF


def migrate_to_0_8_0_from_document_map(catalog, document_map):
    migrate_to_0_8_0_from_docids(catalog, document_map.docid_to_address.keys())


def migrate_to_0_8_0(catalog):
    docids = IF.multiunion([IF.Set(index._indexed()) for index
                            in catalog.values() if hasattr(index, '_indexed')])
    migrate_to_0_8_0_from_docids(catalog, docids)


def migrate_to_0_8_0_from_docids(catalog, docids):
    for index in catalog.values():
        migrate = getattr(index, '_migrate_to_0_8_0', None)
        if migrate is not None:
            migrate(docids)



import unittest

import six

_marker = object()

class TestCatalogTextIndex(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.indexes.text import CatalogTextIndex
        return CatalogTextIndex

    def _makeOne(self, discriminator=_marker):
        def _discriminator(obj, default):
            if obj is _marker:
                return default
            return obj
        if discriminator is _marker:
            discriminator = _discriminator
        return self._getTargetClass()(discriminator)

    def test_class_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyClass
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyClass(ICatalogIndex, self._getTargetClass())

    def test_instance_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyObject
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyObject(ICatalogIndex, self._makeOne())

    def test_class_conforms_to_IIndexSort(self):
        from zope.interface.verify import verifyClass
        from zope.index.interfaces import IIndexSort
        verifyClass(IIndexSort, self._getTargetClass())

    def test_instance_conforms_to_IIndexSort(self):
        from zope.interface.verify import verifyObject
        from zope.index.interfaces import IIndexSort
        verifyObject(IIndexSort, self._makeOne())

    def test_ctor_callback_discriminator(self):
        def _discriminator(obj, default):
            """ """
        index = self._makeOne(_discriminator)
        self.assertTrue(index.discriminator is _discriminator)

    def test_ctor_string_discriminator(self):
        index = self._makeOne('abc')
        self.assertEqual(index.discriminator, 'abc')

    def test_ctor_bad_discriminator(self):
        self.assertRaises(ValueError, self._makeOne, object())

    def test_reindex_doc_doesnt_unindex(self):
        index = self._makeOne()
        index.index_doc(5, 'now is the time')
        index.unindex_doc = lambda *args, **kw: 1/0
        index.reindex_doc(5, 'now is the time')

    def test_sort_no_results(self):
        index = self._makeOne()
        self.assertEqual([], index.sort([]))

    def test_sort_without_weights(self):
        index = self._makeOne()
        self.assertRaises(TypeError, index.sort, [1])

    def test_sort_unlimited_forward(self):
        index = self._makeOne()
        results = {-2: 5.0, 3: 3.0, 0: 4.5}
        expect = [-2, 0, 3]
        self.assertEqual(index.sort(results), expect)

    def test_sort_unlimited_reverse(self):
        index = self._makeOne()
        results = {-2: 5.0, 3: 3.0, 0: 4.5}
        expect = [3, 0, -2]
        self.assertEqual(index.sort(results, reverse=True), expect)

    def test_sort_limited(self):
        index = self._makeOne()
        results = {-2: 5.0, 3: 3.0, 0: 4.5}
        expect = [-2, 0]
        self.assertEqual(index.sort(results, limit=2), expect)

    def test_applyDoesNotContain(self):
        index = self._makeOne()
        index.index_doc(1, u'now is the time')
        index.index_doc(2, u"l'ora \xe9 ora")
        result = sorted(index.applyDoesNotContain('time'))
        self.assertEqual(result, [2])

    def test_applyDoesNotContain_with_unindexed_doc(self):
        def discriminator(obj, default):
            if isinstance(obj, six.string_types):
                return obj
            return default
        index = self._makeOne(discriminator)
        index.index_doc(1, u'now is the time')
        index.index_doc(2, u"l'ora \xe9 ora")
        index.index_doc(3, 3)
        result = sorted(index.applyDoesNotContain('time'))
        self.assertEqual(result, [2, 3])

    def test_applyDoesNotContain_nothing_indexed(self):
        def discriminator(obj, default):
            return default
        index = self._makeOne(discriminator)
        index.index_doc(1, u'now is the time')
        index.index_doc(2, u"l'ora \xe9 ora")
        index.index_doc(3, 3)
        result = sorted(index.applyDoesNotContain('time'))
        self.assertEqual(result, [1, 2, 3])

    def test_docids(self):
        index = self._makeOne()
        index.index_doc(1, u'now is the time')
        index.index_doc(2, u"l'ora \xe9 ora")
        index.index_doc(3, u"you have nice hair.")
        self.assertEqual(set(index.docids()), set((1, 2, 3)))

    def test_unindex_doc_removes_from_docids(self):
        index = self._makeOne()
        index.index_doc(20, _marker)
        self.assertTrue(20 in index.docids())
        index.unindex_doc(20)
        self.assertFalse(20 in index.docids())

    def test_index_doc_then_missing_value(self):
        index = self._makeOne()
        index.index_doc(3, u'Am I rich yet?')
        self.assertEqual(set([3]), set(index.applyContains('rich')))
        self.assertTrue(3 in index.docids())
        index.index_doc(3, _marker)
        self.assertEqual(set(), set(index.applyEq('rich')))
        self.assertTrue(3 in index.docids())

    def test_index_doc_missing_value_then_with_value(self):
        index = self._makeOne()
        index.index_doc(20, _marker)
        self.assertEqual(set(), set(index.applyContains('rich')))
        self.assertTrue(20 in index.docids())
        index.index_doc(20, u'Am I rich yet?')
        self.assertEqual(set([20]), set(index.applyContains('rich')))
        self.assertTrue(20 in index.docids())

    def test_index_doc_missing_value_then_unindex(self):
        index = self._makeOne()
        index.index_doc(20, _marker)
        self.assertEqual(set(), set(index.applyEq('/cmr')))
        self.assertTrue(20 in index.docids())
        index.unindex_doc(20)
        self.assertEqual(set(), set(index.applyEq('/cmr')))
        self.assertFalse(20 in index.docids())

    def test_docids_with_indexed_and_not_indexed(self):
        index = self._makeOne()
        index.index_doc(1, u'Am I rich yet?')
        index.index_doc(2, _marker)
        self.assertEqual(set([1, 2]), set(index.docids()))



from repoze.catalog.catalog import FileStorageCatalogFactory
from repoze.catalog.catalog import ConnectionManager

from repoze.catalog.indexes.field import CatalogFieldIndex
from repoze.catalog.indexes.text import CatalogTextIndex

factory = FileStorageCatalogFactory('catalog.db', 'mycatalog')

_initialized = False

def initialize_catalog():
    global _initialized
    if not _initialized:
        # create a catalog
        manager = ConnectionManager()
        catalog = factory(manager)
        # set up indexes
        catalog['flavors'] = CatalogFieldIndex('flavor')
        catalog['texts'] = CatalogTextIndex('text')
        # commit the indexes
        manager.commit()
        manager.close()
        _initialized = True

class Content(object):
    def __init__(self, flavor, text):
        self.flavor = flavor
        self.text = text

if __name__ == '__main__':
    initialize_catalog()
    manager = ConnectionManager()
    catalog = factory(manager)
    content = {
         1:Content('peach', 'i am so very very peachy'),
         2:Content('pistachio', 'i am nutty'),
         }
    for docid, doc in content.items():
        catalog.index_doc(docid, doc)
    manager.commit()
